"""Defines time dependence"""

from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np
import pydantic.v1 as pydantic

from ..constants import RADIAN
from ..exceptions import SetupError
from .base import Tidy3dBaseModel
from .types import ArrayFloat1D, Ax, PlotVal
from .viz import add_ax_if_none

# in spectrum computation, discard amplitudes with relative magnitude smaller than cutoff
DFT_CUTOFF = 1e-8


class AbstractTimeDependence(ABC, Tidy3dBaseModel):
    """Base class describing time dependence."""

    amplitude: pydantic.NonNegativeFloat = pydantic.Field(
        1.0, title="Amplitude", description="Real-valued maximum amplitude of the time dependence."
    )

    phase: float = pydantic.Field(
        0.0, title="Phase", description="Phase shift of the time dependence.", units=RADIAN
    )

    @abstractmethod
    def amp_time(self, time: float) -> complex:
        """Complex-valued amplitude as a function of time.

        Parameters
        ----------
        time : float
            Time in seconds.

        Returns
        -------
        complex
            Complex-valued amplitude at that time.
        """

    def spectrum(
        self,
        times: ArrayFloat1D,
        freqs: ArrayFloat1D,
        dt: float,
    ) -> complex:
        """Complex-valued spectrum as a function of frequency.
        Note: Only the real part of the time signal is used.

        Parameters
        ----------
        times : np.ndarray
            Times to use to evaluate spectrum Fourier transform.
            (Typically the simulation time mesh).
        freqs : np.ndarray
            Frequencies in Hz to evaluate spectrum at.
        dt : float or np.ndarray
            Time step to weight FT integral with.
            If array, use to weigh each of the time intervals in ``times``.

        Returns
        -------
        np.ndarray
            Complex-valued array (of len(freqs)) containing spectrum at those frequencies.
        """

        times = np.array(times)
        freqs = np.array(freqs)
        time_amps = np.real(self.amp_time(times))

        # if all time amplitudes are zero, just return (complex-valued) zeros for spectrum
        if np.all(np.equal(time_amps, 0.0)):
            return (0.0 + 0.0j) * np.zeros_like(freqs)

        # Cut to only relevant times
        relevant_time_inds = np.where(np.abs(time_amps) / np.amax(np.abs(time_amps)) > DFT_CUTOFF)
        # find first and last index where the filter is True
        start_ind = relevant_time_inds[0][0]
        stop_ind = relevant_time_inds[0][-1] + 1
        time_amps = time_amps[start_ind:stop_ind]
        times_cut = times[start_ind:stop_ind]
        if times_cut.size == 0:
            return (0.0 + 0.0j) * np.zeros_like(freqs)

        # only need to compute DTFT kernel for distinct dts
        # usually, there is only one dt, if times is simulation time mesh
        dts = np.diff(times_cut)
        dts_unique, kernel_indices = np.unique(dts, return_inverse=True)

        dft_kernels = [np.exp(2j * np.pi * freqs * curr_dt) for curr_dt in dts_unique]
        running_kernel = np.exp(2j * np.pi * freqs * times_cut[0])
        dft = np.zeros(len(freqs), dtype=complex)
        for amp, kernel_index in zip(time_amps, kernel_indices):
            dft += running_kernel * amp
            running_kernel *= dft_kernels[kernel_index]

        # kernel_indices was one index shorter than time_amps
        dft += running_kernel * time_amps[-1]

        return dt * dft / np.sqrt(2 * np.pi)

    @add_ax_if_none
    def plot_spectrum_in_frequency_range(
        self,
        times: ArrayFloat1D,
        fmin: float,
        fmax: float,
        num_freqs: int = 101,
        val: PlotVal = "real",
        ax: Ax = None,
    ) -> Ax:
        """Plot the complex-valued amplitude of the time-dependence.
        Note: Only the real part of the time signal is used.

        Parameters
        ----------
        times : np.ndarray
            Array of evenly-spaced times (seconds) to evaluate time-dependence at.
            The spectrum is computed from this value and the time frequency content.
            To see spectrum for a specific :class:`Simulation`,
            pass ``simulation.tmesh``.
        fmin : float
            Lower bound of frequency for the spectrum plot.
        fmax : float
            Upper bound of frequency for the spectrum plot.
        num_freqs : int = 101
            Number of frequencies to plot within the [fmin, fmax].
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        times = np.array(times)

        dts = np.diff(times)
        if not np.allclose(dts, dts[0] * np.ones_like(dts), atol=1e-17):
            raise SetupError("Supplied times not evenly spaced.")

        dt = np.mean(dts)
        freqs = np.linspace(fmin, fmax, num_freqs)

        spectrum = self.spectrum(times=times, dt=dt, freqs=freqs)

        if val == "real":
            ax.plot(freqs, spectrum.real, color="blueviolet", label="real")
        elif val == "imag":
            ax.plot(freqs, spectrum.imag, color="crimson", label="imag")
        elif val == "abs":
            ax.plot(freqs, np.abs(spectrum), color="k", label="abs")
        else:
            raise ValueError(f"Plot 'val' option of '{val}' not recognized.")
        ax.set_xlabel("frequency (Hz)")
        ax.set_title("source spectrum")
        ax.legend()
        ax.set_aspect("auto")
        return ax

    @add_ax_if_none
    def plot(self, times: ArrayFloat1D, val: PlotVal = "real", ax: Ax = None) -> Ax:
        """Plot the complex-valued amplitude of the time-dependence.

        Parameters
        ----------
        times : np.ndarray
            Array of times (seconds) to plot source at.
            To see source time amplitude for a specific :class:`Simulation`,
            pass ``simulation.tmesh``.
        val : Literal['real', 'imag', 'abs'] = 'real'
            Which part of the spectrum to plot.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        times = np.array(times)
        amp_complex = self.amp_time(times)

        if val == "real":
            ax.plot(times, amp_complex.real, color="blueviolet", label="real")
        elif val == "imag":
            ax.plot(times, amp_complex.imag, color="crimson", label="imag")
        elif val == "abs":
            ax.plot(times, np.abs(amp_complex), color="k", label="abs")
        else:
            raise ValueError(f"Plot 'val' option of '{val}' not recognized.")
        ax.set_xlabel("time (s)")
        ax.set_title("source amplitude")
        ax.legend()
        ax.set_aspect("auto")
        return ax



"""Collection of functions for automatically generating a nonuniform grid."""

import warnings
from abc import ABC, abstractmethod
from itertools import compress
from math import isclose
from typing import Dict, List, Tuple, Union

import numpy as np
import pydantic.v1 as pd
from pyroots import Brentq
from shapely.errors import ShapelyDeprecationWarning
from shapely.geometry import box as shapely_box
from shapely.strtree import STRtree

from ...constants import C_0, fp_eps
from ...exceptions import SetupError, ValidationError
from ...log import log
from ..base import Tidy3dBaseModel
from ..structure import MeshOverrideStructure, Structure, StructureType
from ..types import ArrayFloat1D, Axis, Bound, CoordinateOptional

_ROOTS_TOL = 1e-10

# Shrink min_step a little so that if e.g. a structure has target dl = 0.1 and a width of 0.1,
# a grid point will be added on both sides of the structure. Without this factor, the mesher
# ``is_close`` check will deem that the second point is too close.
MIN_STEP_SCALE = 0.9999


class Mesher(Tidy3dBaseModel, ABC):
    """Abstract class for automatic meshing."""

    @abstractmethod
    def parse_structures(
        self,
        axis: Axis,
        structures: List[StructureType],
        wavelength: pd.PositiveFloat,
        min_steps_per_wvl: pd.NonNegativeInt,
        dl_min: pd.NonNegativeFloat,
        dl_max: pd.NonNegativeFloat,
    ) -> Tuple[ArrayFloat1D, ArrayFloat1D]:
        """Calculate the positions of all bounding box interfaces along a given axis."""

    @abstractmethod
    def insert_snapping_points(
        self,
        dl_min: pd.NonNegativeFloat,
        axis: Axis,
        interval_coords: ArrayFloat1D,
        max_dl_list: ArrayFloat1D,
        snapping_points: List[CoordinateOptional],
    ) -> Tuple[ArrayFloat1D, ArrayFloat1D]:
        """Insert snapping_points to the intervals."""

    @abstractmethod
    def make_grid_multiple_intervals(
        self,
        max_dl_list: ArrayFloat1D,
        len_interval_list: ArrayFloat1D,
        max_scale: float,
        is_periodic: bool,
    ) -> List[ArrayFloat1D]:
        """Create grid steps in multiple connecting intervals."""

    @staticmethod
    def structure_step(
        structure: Structure,
        wavelength: float,
        min_steps_per_wvl: float,
    ) -> ArrayFloat1D:
        """Get the minimum mesh required in a :class:`.Structure`. Special media are set to index of 1,
        which would usually make the medium ignored in the meshing, while the structure geometry
        would still be used to place grid boundaries.

        Parameters
        ----------
        structure : :class:`.Structure`
            A regular structure
        wavelength : float
            Wavelength to use for the step size and for dispersive media epsilon.
        min_steps_per_wvl : float
            Minimum requested steps per wavelength.
        """

    @staticmethod
    def make_shapely_box(bbox: Bound) -> shapely_box:
        """Make a shapely box out of bounds."""
        return shapely_box(bbox[0, 0], bbox[0, 1], bbox[1, 0], bbox[1, 1])


class GradedMesher(Mesher):
    """Implements automatic nonuniform meshing with a set minimum steps per wavelength and
    a graded mesh expanding from higher- to lower-resolution regions."""

    def insert_snapping_points(
        self,
        dl_min: pd.NonNegativeFloat,
        axis: Axis,
        interval_coords: ArrayFloat1D,
        max_dl_list: ArrayFloat1D,
        snapping_points: List[CoordinateOptional],
    ) -> Tuple[ArrayFloat1D, ArrayFloat1D]:
        """Insert snapping_points to the intervals.

        Parameters
        ----------
        dl_min: pd.NonNegativeFloat
            Lower bound of grid size.
        axis : Axis
            Axis index along which to operate.
        interval_coords : ArrayFloat1D
            Coordinate of interval boundaries.
        max_dl_list : ArrayFloat1D
            Maximal allowed step size of each interval generated from `parse_structures`.
        snapping_points : List[CoordinateOptional]
            A set of points that enforce grid boundaries to pass through them.

        Returns
        -------
        interval_coords : Array
            An array of coordinates, where the first element is the simulation min boundary, the
            last element is the simulation max boundary, and the intermediate coordinates are all
            locations with a snapping_point or a structure has a bounding box edge along the
            specified axis. The boundaries are filtered such that no interval is smaller than the
            smallest of the ``max_steps``.
        max_steps : array_like
            An array of size ``interval_coords.size - 1`` giving the maximum grid step required in
            each ``interval_coords[i]:interval_coords[i+1]`` interval, depending on the materials
            in that interval, the supplied wavelength, and the minimum required step per wavelength.

        """
        # Don't do anything for a 2D-like simulation, or no snapping points
        if interval_coords.size == 1 or len(snapping_points) < 1:
            return interval_coords, max_dl_list

        # If the user has set dl_min, use that as the min_step
        if dl_min != 0.0:
            min_step = dl_min
        else:
            min_step = np.amin(max_dl_list) / 2

        for point in snapping_points:
            new_coord = point[axis]
            if new_coord is None:
                continue
            # Skip if the point is outside the domain
            if new_coord >= interval_coords[-1] or new_coord <= interval_coords[0]:
                continue
            # search insertion location
            ind = np.searchsorted(interval_coords, new_coord, side="left")
            d_1 = abs(new_coord - interval_coords[ind - 1])
            d_2 = abs(new_coord - interval_coords[ind])
            # Skip snapping_points if the distances to existing interval boundaries are
            # smaller than `min_step` defined above.
            if d_1 < min_step and d_2 < min_step:
                continue
            # or, if the snapping point is near the first interval boundary,
            # give priority to the snapping_point and replace
            elif d_1 < min_step:
                # Don't replace if the boundary is the simulation boundary
                if ind == 0:
                    continue
                interval_coords[ind - 1] = new_coord
                continue
            # or, if the snapping point is near the second interval boundary,
            # give priority to the snapping_point and replace
            elif d_2 < min_step:
                # Don't replace if the boundary is the simulation boundary
                if ind == len(interval_coords) - 1:
                    continue
                interval_coords[ind] = new_coord
                continue
            # otherwise add the snapping point directly since there is sufficient space
            interval_coords = np.insert(interval_coords, ind, new_coord)
            max_dl_list = np.insert(max_dl_list, ind - 1, max_dl_list[ind - 1])
        return interval_coords, max_dl_list

    def parse_structures(
        self,
        axis: Axis,
        structures: List[StructureType],
        wavelength: pd.PositiveFloat,
        min_steps_per_wvl: pd.NonNegativeInt,
        dl_min: pd.NonNegativeFloat,
        dl_max: pd.NonNegativeFloat,
    ) -> Tuple[ArrayFloat1D, ArrayFloat1D]:
        """Calculate the positions of all bounding box interfaces along a given axis.
        In this implementation, in most cases the complexity should be O(len(structures)**2),
        although the worst-case complexity may approach O(len(structures)**3).
        However this should only happen in some very contrived cases.

        Parameters
        ----------
        axis : Axis
            Axis index along which to operate.
        structures : List[StructureType]
            List of structures, with the simulation structure being the first item.
        wavelength : pd.PositiveFloat
            Wavelength to use for the step size and for dispersive media epsilon.
        min_steps_per_wvl : pd.NonNegativeInt
            Minimum requested steps per wavelength.
        dl_min: pd.NonNegativeFloat
            Lower bound of grid size.
        dl_max: pd.NonNegativeFloat
            Upper bound of grid size.

        Returns
        -------
        interval_coords : Array
            An array of coordinates, where the first element is the simulation min boundary, the
            last element is the simulation max boundary, and the intermediate coordinates are all
            locations where a structure has a bounding box edge along the specified axis.
            The boundaries are filtered such that no interval is smaller than the smallest
            of the ``max_steps``.
        max_steps : array_like
            An array of size ``interval_coords.size - 1`` giving the maximum grid step required in
            each ``interval_coords[i]:interval_coords[i+1]`` interval, depending on the materials
            in that interval, the supplied wavelength, and the minimum required step per wavelength.
        """

        # Simulation boundaries
        sim_bmin, sim_bmax = structures[0].geometry.bounds
        domain_bounds = np.array([sim_bmin[axis], sim_bmax[axis]])

        # For MeshOverrideStructure, we allow ``dl`` along some axis to be ``None``
        # so that no override occurs along this axis. So first, we filter all those
        # structures.
        structures_effective = self.filter_structures_effective_dl(structures, axis)

        # Special attention needs to be paid to enforced or unshadowed overrideStructures.
        # 1) enforced structures shouldn't be overridden by other structures; for overlapping enforced
        #    structures, the grid size of the overlapped region is determined by the last enforced structure.
        # 2) unshadowed structures don't override other structures based on the structure list order,
        #    but by grid size; additionally, unshadowed structures won't add new intervals if they
        #    don't reduce grid size in the intervals.
        #  We take two steps to implement the feature:
        # 1) reorder structure list so that enforce = True structures are shifted to
        #    the end of the last, and unshadowed structures to the begining (but after simulation structure).
        # 2) in each interval, the maximal grid size is:
        #    a) no enforced structure: min(grid size of each structure).
        #    b) with enforced structure: grid size of the last override structure.

        # reorder structure list
        num_unenforced, structures_ordered = self.reorder_structures(structures_effective)

        # no containment check for unshadowed structures
        skip_containment = [
            isinstance(structure, MeshOverrideStructure) and not structure.shadow
            for structure in structures_ordered
        ]

        # Required maximum steps in every structure
        structure_steps = self.structure_steps(
            structures_ordered, wavelength, min_steps_per_wvl, dl_min, dl_max, axis
        )
        # Smallest of the maximum steps
        min_step = MIN_STEP_SCALE * np.amin(structure_steps)

        # If empty simulation, return
        if len(structures) == 1:
            interval_coords, max_steps = self.filter_min_step(domain_bounds, structure_steps)
            return np.array(interval_coords), np.array(max_steps)

        # Bounding boxes with the meshing axis rotated to z
        struct_bbox = self.rotate_structure_bounds(structures_ordered, axis)
        # Rtree from the 2D part of the bounding boxes
        tree = self.bounds_2d_tree(struct_bbox)

        intervals = {"coords": list(domain_bounds), "structs": [[]]}
        """ Build the ``intervals`` dictionary. ``intervals["coords"]`` gets populated based on the
        bounding boxes of all structures in the list (some filtering is done to exclude points that
        will be too close together compared to the absolute lower desired step). At every point, the
        ``"structs"`` list has length one lower than the ``"coords"`` list, and each element is
        another list of all structure indexes that are found in the interval formed by ``coords[i]``
        and ``coords[i + 1]``. This only includes structures that have a physical presence in the
        interval, i.e. it excludes structures that are completely covered by higher-up ones.

        To build this, we iterate in reverse order as latter structures override earlier ones.
        We also handle containment in the following way (note - we work with bounding boxes only):
         1. If a structure that is lower in the list than the current structure is found to be
          completely contained in 3D, the corresponding ``struct_bbox`` is immediately set to
          ``None`` and nothing more will be done using that structure.
         2. If the current structure is covering an interval but there's a higher-up structure that
          contains it in 2D and also covers the same interval, then it will not be added to the
          ``intervals["structs"] list for that interval.
         3. If the current structure is found to not cover any interval, its bounding box
          is set to ``None``, so that it will not affect structures that lie below it as per point
          2. A warning is also raised since the structure will have an unpredictable effect on the
          material coefficients used in the simulation.
        """

        with log:
            for str_ind in range(len(structures_ordered) - 1, -1, -1):
                # 3D and 2D bounding box of current structure
                bbox = struct_bbox[str_ind]
                if bbox is None:
                    # Structure has been removed because it is completely contained
                    continue

                query_inds = []
                if not skip_containment[str_ind]:
                    bbox_2d = self.make_shapely_box(bbox)

                    # List of structure indexes that may intersect the current structure in 2D
                    try:
                        query_inds = tree.query_items(bbox_2d)
                    except AttributeError:
                        query_inds = tree.query(bbox_2d)

                # Remove all lower structures that the current structure completely contains
                inds_lower = [
                    ind
                    for ind in query_inds
                    if ind < str_ind and struct_bbox[ind] is not None and not skip_containment[ind]
                ]
                query_bbox = [struct_bbox[ind] for ind in inds_lower]
                bbox_contains_inds = self.contains_3d(bbox, query_bbox)
                for ind in bbox_contains_inds:
                    struct_bbox[inds_lower[ind]] = None

                # List of structure bboxes that contain the current structure in 2D
                bbox_contained_2d = []
                if not skip_containment[str_ind]:
                    inds_upper = [ind for ind in query_inds if ind > str_ind]
                    query_bbox = [
                        struct_bbox[ind]
                        for ind in inds_upper
                        if struct_bbox[ind] is not None and not skip_containment[ind]
                    ]
                    bbox_contained_2d = self.contained_2d(bbox, query_bbox)

                # Handle insertion of the current structure bounds in the intervals
                # The intervals list is modified in-place
                too_small = self.insert_bbox(
                    intervals,
                    str_ind,
                    bbox,
                    bbox_contained_2d,
                    min_step,
                    structure_steps,
                    skip_containment[str_ind],
                )
                if too_small and (bbox[1, 2] - bbox[0, 2]) > 0:
                    # If the structure is too small (but not 0D), issue a warning
                    log.warning(
                        f"A structure has a nonzero dimension along axis {'xyz'[axis]}, which "
                        "is however too small compared to the generated mesh step along that "
                        "direction. This could produce unpredictable results. We recommend "
                        "increasing the resolution, or adding a mesh override structure to ensure "
                        "that all geometries are at least one pixel thick along all dimensions."
                    )
                    # Also remove this structure from the bbox list so that lower structures
                    # will not be affected by it, as it was not added to any interval.
                    struct_bbox[str_ind] = None

        # Truncate intervals to domain bounds
        coords = np.array(intervals["coords"])
        num_ints = len(intervals["structs"])
        in_domain = np.nonzero((coords >= domain_bounds[0]) * (coords <= domain_bounds[1]))[0]
        intervals["coords"] = [intervals["coords"][int(i)] for i in in_domain]
        intervals["structs"] = [intervals["structs"][int(i)] for i in in_domain if i < num_ints]

        # Compute the maximum allowed step size in each interval
        max_steps = []
        for coord_ind, _ in enumerate(intervals["coords"][:-1]):
            # if there are any enforced structure in the interval, use the last structure
            if max(intervals["structs"][coord_ind]) >= num_unenforced:
                max_step = structure_steps[max(intervals["structs"][coord_ind])]
                max_steps.append(max_step)
            # otherwise, define the max step as the minimum over all medium steps
            # of media in this interval
            else:
                max_step = np.amin(structure_steps[intervals["structs"][coord_ind]])
                max_steps.append(max_step)

        # Re-evaluate the absolute smallest min_step and remove intervals that are smaller than that
        intervals["coords"], max_steps = self.filter_min_step(intervals["coords"], max_steps)

        return np.array(intervals["coords"]), np.array(max_steps)

    def insert_bbox(
        self,
        intervals: Dict[str, List],
        str_ind: int,
        str_bbox: ArrayFloat1D,
        bbox_contained_2d: List[ArrayFloat1D],
        min_step: float,
        structure_steps: ArrayFloat1D,
        unshadowed: bool,
    ) -> Dict[str, List]:
        """Figure out where to place the bounding box coordinates of current structure.
        For both the left and the right bounds of the structure along the meshing direction,
        we check if they are not too close to an already existing coordinate, if the
        structure is not completely covered by another structure at that location, and if
        the structure is unshadowed and it refines grid size,
        Only then we add that boundary to the list of interval coordinates.
        We also don't add the bounds if ``str_ind==0``, since the domain bounds have already
        been added to the interval coords at the start.
        We also compute ``indmin`` and ``indmax`` indexes into the list of intervals, such that
        the current structure is added to all intervals in the range (indmin, indmax).

        Parameters
        ----------
        intervals : Dict[str, List]
            Dictionary containing the coordinates of the interval boundaries, and a list
            of lists of structures contained in each interval.
        str_ind : int
            Index of the current structure.
        str_bbox : ArrayFloat1D
            Bounding box of the current structure.
        bbox_contained_2d : List[ArrayFloat1D]
            List of 3D bounding boxes that contain the current structure in 2D.
        min_step : float
            Absolute minimum interval size to impose.
        structure_steps : ArrayFloat1D
            Maximal grid size in each structure.
        unshadowed : Bool
            Whether the current structure is unshadowed.

        Returns
        -------
        structure_too_small : bool
            True if the structure did not span any interval after coordinates were inserted. This
            would happen if the structure size is too small compared to the minimum step.

        Note
        ----
            This function modifies ``intervals`` in-place.
        """

        coords = intervals["coords"]
        structs = intervals["structs"]

        min_step_check = MIN_STEP_SCALE * min_step

        # Left structure bound
        bound_coord = str_bbox[0, 2]
        indsmin = np.nonzero(bound_coord <= coords)[0]
        indmin = int(indsmin[0])  # coordinate is in interval index ``indmin - 1``
        is_close_l = self.is_close(bound_coord, coords, indmin - 1, min_step_check)
        is_close_r = self.is_close(bound_coord, coords, indmin, min_step_check)
        is_contained = self.is_contained(bound_coord, bbox_contained_2d)

        # special treatment to unshadowed structure
        skip_unshadowed = False
        if unshadowed and indmin > 0:
            grid_size_str = structure_steps[str_ind]
            min_grid_size = min(
                (structure_steps[ind] for ind in structs[indmin - 1]), default=grid_size_str
            )
            if not (isclose(grid_size_str, min_grid_size) or grid_size_str < min_grid_size):
                skip_unshadowed = True

        # Decide on whether coordinate should be inserted or indmin modified
        if is_close_l or skip_unshadowed:
            # Don't insert coordinate but decrease indmin.
            # Note that for `skip_unshadowed`, it makes no difference to decrease indmin or not, as it will not
            # decrease grid size in this interval.
            indmin -= 1
        elif not is_close_r and not is_contained and str_ind > 0:
            # Add current structure bounding box coordinates
            coords.insert(indmin, bound_coord)
            # Copy the structure containment list to the newly created interval
            struct_list = structs[max(0, indmin - 1)]
            structs.insert(indmin, struct_list.copy())

        # Right structure bound
        bound_coord = str_bbox[1, 2]
        indsmax = np.nonzero(bound_coord >= coords)[0]
        indmax = int(indsmax[-1])  # coordinate is in interval index ``indmax``
        is_close_l = self.is_close(bound_coord, coords, indmax, min_step_check)
        is_close_r = self.is_close(bound_coord, coords, indmax + 1, min_step_check)
        is_contained = self.is_contained(bound_coord, bbox_contained_2d)

        # special treatment to unshadowed structure
        skip_unshadowed = False
        if unshadowed and indmax < len(structs):
            grid_size_str = structure_steps[str_ind]
            min_grid_size = min(
                (structure_steps[ind] for ind in structs[indmax]), default=grid_size_str
            )
            if not (isclose(grid_size_str, min_grid_size) or grid_size_str < min_grid_size):
                skip_unshadowed = True

        # Decide on whether coordinate should be inserted or indmax modified
        if is_close_r or skip_unshadowed:
            # Don't insert coordinate but increase indmax
            indmax += 1
        elif not is_close_l and not is_contained and str_ind > 0:
            indmax += 1
            # Add current structure bounding box coordinates
            coords.insert(indmax, bound_coord)
            # Copy the structure containment list to the newly created interval
            struct_list = structs[min(indmax - 1, len(structs) - 1)]
            structs.insert(indmax, struct_list.copy())

        # Add the current structure index to all intervals that it spans, if it is not
        # contained in any of the latter structures
        for interval_ind in range(indmin, indmax):
            # Check at the midpoint to avoid numerical issues at the interval boundaries
            mid_coord = (coords[interval_ind] + coords[interval_ind + 1]) / 2
            if not self.is_contained(mid_coord, bbox_contained_2d):
                structs[interval_ind].append(str_ind)

        return indmin >= indmax

    @staticmethod
    def reorder_structures(
        structures: List[StructureType],
    ) -> Tuple[int, List[StructureType]]:
        """Reorder structure list to order as follows:
        1). simulation structure `str[0]` remains as the first structure;
        2). MeshOverrideStructures with ``shadow=False``;
        3). Structures other than 1),2),4);
        4). MeshOverrideStructures with ``enforce=True``.

        Parameters
        ----------
        structures : List[StructureType]
            List of structures, with the simulation structure being the first item.

        Returns
        -------
        Tuple[int, List[StructureType]]
            The number of unenforced structures, reordered structure list

        """

        # boolean list for enforced structures
        enforced_list = [
            isinstance(structure, MeshOverrideStructure) and structure.enforce
            for structure in structures
        ]
        # boolean list for unshadowed structures
        unshadowed_list = [
            isinstance(structure, MeshOverrideStructure) and not structure.shadow
            for structure in structures
        ]
        # if no special structure, a quick return here
        if not (any(enforced_list) or any(unshadowed_list)):
            return len(structures), structures

        # filter structures
        structures_enforced = list(compress(structures, enforced_list))
        structures_unshadowed = list(compress(structures, unshadowed_list))
        structures_others = list(
            compress(
                structures,
                [
                    not (enforced or unshadowed)
                    for unshadowed, enforced in zip(unshadowed_list, enforced_list)
                ],
            )
        )

        ordered_structures = (
            [structures_others[0]]
            + structures_unshadowed
            + structures_others[1:]
            + structures_enforced
        )
        return len(structures_others) + len(structures_unshadowed), ordered_structures

    @staticmethod
    def filter_structures_effective_dl(
        structures: List[StructureType], axis: Axis
    ) -> List[StructureType]:
        """For :class:`.MeshOverrideStructure`, we allow ``dl`` along some axis
        to be ``None`` so that no override occurs along this axis.Here those
        structures with ``dl[axis]=None`` is filtered.

        Parameters
        ----------
        structures : List[StructureType]
            List of structures, with the simulation structure being the first item.
        axis : Axis
            Axis index to place last.

        Returns
        -------
        List[StructureType]
            A list of filtered structures whose ``dl`` along this axis is not ``None``.
        """

        structures_filtered = []
        for structure in structures:
            if isinstance(structure, MeshOverrideStructure):
                # skip override structure if dl not defined along axis
                if structure.dl[axis] is None:
                    continue
            structures_filtered.append(structure)
        return structures_filtered

    @staticmethod
    def structure_step(
        structure: Structure,
        wavelength: float,
        min_steps_per_wvl: float,
    ) -> ArrayFloat1D:
        """Get the minimum mesh required in a :class:`.Structure`. Special media are set to index of 1,
        which would usually make the medium ignored in the meshing, while the structure geometry
        would still be used to place grid boundaries.

        Parameters
        ----------
        structure : :class:`.Structure`
            A regular structure
        wavelength : float
            Wavelength to use for the step size and for dispersive media epsilon.
        min_steps_per_wvl : float
            Minimum requested steps per wavelength.
        """

        eps_diagonal = structure.medium.eps_diagonal_numerical(C_0 / wavelength)
        n, k = structure.medium.eps_complex_to_nk(eps_diagonal)
        # take max among all directions because perpendicular eps defines wavelength
        max_n_abs = np.max(np.abs(n))
        max_k_abs = np.max(np.abs(k))
        index = np.max([max_n_abs, max_k_abs])
        return wavelength / index / min_steps_per_wvl

    @staticmethod
    def structure_steps(
        structures: List[StructureType],
        wavelength: float,
        min_steps_per_wvl: float,
        dl_min: pd.NonNegativeFloat,
        dl_max: pd.NonNegativeFloat,
        axis: Axis,
    ) -> ArrayFloat1D:
        """Get the minimum mesh required in each structure. Special media are set to index of 1,
        which would usually make the medium ignored in the meshing, while the structure geometry
        would still be used to place grid boundaries.

        Parameters
        ----------
        structures : List[Structure]
            List of structures, with the simulation structure being the first item.
        wavelength : float
            Wavelength to use for the step size and for dispersive media epsilon.
        min_steps_per_wvl : float
            Minimum requested steps per wavelength.
        dl_min: pd.NonNegativeFloat
            Lower bound of grid size.
        dl_max: pd.NonNegativeFloat
            Upper bound of grid size.
        axis : Axis
            Axis index along which to operate.
        """
        min_steps = []
        for structure in structures:
            if isinstance(structure, Structure):
                min_steps.append(
                    GradedMesher.structure_step(structure, wavelength, min_steps_per_wvl)
                )
            elif isinstance(structure, MeshOverrideStructure):
                min_steps.append(structure.dl[axis])
        min_steps = np.array(min_steps)
        min_steps = np.where(min_steps > dl_max, dl_max, min_steps)
        return np.where(min_steps < dl_min, dl_min, min_steps)

    @staticmethod
    def rotate_structure_bounds(structures: List[StructureType], axis: Axis) -> List[ArrayFloat1D]:
        """Get structure bounding boxes with a given ``axis`` rotated to z.

        Parameters
        ----------
        structures : List[StructureType]
            List of structures, with the simulation structure being the first item.
        axis : Axis
            Axis index to place last.

        Returns
        -------
        List[ArrayFloat1D]
            A list of the bounding boxes of shape ``(2, 3)`` for each structure, with the bounds
            along ``axis`` being ``(:, 2)``.
        """
        struct_bbox = []
        for structure in structures:
            # Get 3D bounding box and rotate axes
            bmin, bmax = structure.geometry.bounds
            bmin_ax, bmin_plane = structure.geometry.pop_axis(bmin, axis=axis)
            bmax_ax, bmax_plane = structure.geometry.pop_axis(bmax, axis=axis)
            bounds = np.array([list(bmin_plane) + [bmin_ax], list(bmax_plane) + [bmax_ax]])
            struct_bbox.append(bounds)
        return struct_bbox

    @staticmethod
    def bounds_2d_tree(struct_bbox: List[ArrayFloat1D]):
        """Make a shapely Rtree for the 2D bounding boxes of all structures in the plane
        perpendicular to the meshing axis."""

        boxes_2d = []
        for bbox in struct_bbox:
            box = Mesher.make_shapely_box(bbox)
            boxes_2d.append(box)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=ShapelyDeprecationWarning)
            stree = STRtree(boxes_2d)
        return stree

    @staticmethod
    def contained_2d(bbox0: ArrayFloat1D, query_bbox: List[ArrayFloat1D]) -> List[ArrayFloat1D]:
        """Return a list of all bounding boxes among ``query_bbox`` that contain ``bbox0`` in 2D."""
        return [
            bbox
            for bbox in query_bbox
            if all(
                [
                    bbox0[0, 0] + fp_eps >= bbox[0, 0],
                    bbox0[1, 0] <= bbox[1, 0] + fp_eps,
                    bbox0[0, 1] + fp_eps >= bbox[0, 1],
                    bbox0[1, 1] <= bbox[1, 1] + fp_eps,
                ]
            )
        ]

    @staticmethod
    def contains_3d(bbox0: ArrayFloat1D, query_bbox: List[ArrayFloat1D]) -> List[int]:
        """Return a list of all indexes of bounding boxes in the ``query_bbox`` list that ``bbox0``
        fully contains."""
        return [
            ind
            for ind, bbox in enumerate(query_bbox)
            if all(
                [
                    bbox[0, 0] + fp_eps >= bbox0[0, 0],
                    bbox[1, 0] <= bbox0[1, 0] + fp_eps,
                    bbox[0, 1] + fp_eps >= bbox0[0, 1],
                    bbox[1, 1] <= bbox0[1, 1] + fp_eps,
                    bbox[0, 2] + fp_eps >= bbox0[0, 2],
                    bbox[1, 2] <= bbox0[1, 2] + fp_eps,
                ]
            )
        ]

    @staticmethod
    def is_close(coord: float, interval_coords: List[float], coord_ind: int, atol: float) -> bool:
        """Check if a given ``coord`` is within ``atol`` of an interval coordinate at a given
        interval index. If the index is out of bounds, return ``False``."""
        return (
            isclose(coord, interval_coords[coord_ind], abs_tol=atol)
            if 0 <= coord_ind < len(interval_coords)
            else False
        )

    @staticmethod
    def is_contained(normal_pos: float, contained_2d: List[ArrayFloat1D]) -> bool:
        """Check if a given ``normal_pos`` along the meshing direction is contained inside any
        of the bounding boxes that are in the ``contained_2d`` list.
        """
        return any(
            contain_box[0, 2] <= normal_pos <= contain_box[1, 2] for contain_box in contained_2d
        )

    @staticmethod
    def filter_min_step(
        interval_coords: List[float], max_steps: List[float]
    ) -> Tuple[List[float], List[float]]:
        """Filter intervals that are smaller than the absolute smallest of the ``max_steps``."""

        # Re-compute minimum step in case some high-index structures were completely covered
        min_step_check = MIN_STEP_SCALE * np.amin(max_steps)

        # Filter interval coordinates and max_steps
        coords_filter = [interval_coords[0]]
        steps_filter = []
        for coord_ind, coord in enumerate(interval_coords[1:]):
            if coord - coords_filter[-1] >= min_step_check or (
                coord_ind == len(interval_coords) - 2 and not isclose(coord, coords_filter[-1])
            ):
                coords_filter.append(coord)
                steps_filter.append(max_steps[coord_ind])

        return coords_filter, steps_filter

    def make_grid_multiple_intervals(
        self,
        max_dl_list: ArrayFloat1D,
        len_interval_list: ArrayFloat1D,
        max_scale: float,
        is_periodic: bool,
    ) -> List[ArrayFloat1D]:
        """Create grid steps in multiple connecting intervals of length specified by
        ``len_interval_list``. The maximal allowed step size in each interval is given by
        ``max_dl_list``. The maximum ratio between neighboring steps is bounded by ``max_scale``.

        Parameters
        ----------
        max_dl_list : ArrayFloat1D
            Maximal allowed step size of each interval.
        len_interval_list : ArrayFloat1D
            A list of interval lengths
        max_scale : float
            Maximal ratio between consecutive steps.
        is_periodic : bool
            Apply periodic boundary condition or not.

        Returns
        -------
        List[ArrayFloat1D]
            A list of of step sizes in each interval.
        """

        num_intervals = len(len_interval_list)
        if len(max_dl_list) != num_intervals:
            raise SetupError(
                "Maximal step size list should have the same length as len_interval_list."
            )

        # initialize step size on the left and right boundary of each interval
        # by assuming possible non-integar step number
        left_dl_list, right_dl_list = self.grid_multiple_interval_analy_refinement(
            max_dl_list, len_interval_list, max_scale, is_periodic
        )

        # initialize grid steps
        dl_list = [
            self.make_grid_in_interval(
                left_dl_list[interval_ind],
                right_dl_list[interval_ind],
                max_dl_list[interval_ind],
                max_scale,
                len_interval_list[interval_ind],
            )
            for interval_ind in range(num_intervals)
        ]

        # refinement
        refine_edge = 1

        while refine_edge > 0:
            refine_edge = 0
            for interval_ind in range(num_intervals):
                # the step size on the left and right boundary
                left_dl = dl_list[interval_ind][0]
                right_dl = dl_list[interval_ind][-1]
                # the step size to the left and right boundary (neighbor interval)
                left_neighbor_dl = dl_list[interval_ind - 1][-1]
                right_neighbor_dl = dl_list[(interval_ind + 1) % num_intervals][0]

                # for non-periodic case
                if not is_periodic:
                    if interval_ind == 0:
                        left_neighbor_dl = left_dl
                    if interval_ind == num_intervals - 1:
                        right_neighbor_dl = right_dl

                # compare to the neighbor
                refine_local = 0
                if left_dl / left_neighbor_dl > max_scale:
                    left_dl = left_neighbor_dl * (max_scale - fp_eps)
                    refine_edge += 1
                    refine_local += 1

                if right_dl / right_neighbor_dl > max_scale:
                    right_dl = right_neighbor_dl * (max_scale - fp_eps)
                    refine_edge += 1
                    refine_local += 1

                # update grid steps in this interval if necessary
                if refine_local > 0:
                    dl_list[interval_ind] = self.make_grid_in_interval(
                        left_dl,
                        right_dl,
                        max_dl_list[interval_ind],
                        max_scale,
                        len_interval_list[interval_ind],
                    )

        return dl_list

    def grid_multiple_interval_analy_refinement(
        self,
        max_dl_list: ArrayFloat1D,
        len_interval_list: ArrayFloat1D,
        max_scale: float,
        is_periodic: bool,
    ) -> Tuple[ArrayFloat1D, ArrayFloat1D]:
        """Analytical refinement for multiple intervals. "analytical" meaning we allow
        non-integar step sizes, so that we don't consider snapping here.

        Parameters
        ----------
        max_dl_list : ArrayFloat1D
            Maximal allowed step size of each interval.
        len_interval_list : ArrayFloat1D
            A list of interval lengths
        max_scale : float
            Maximal ratio between consecutive steps.
        is_periodic : bool
            Apply periodic boundary condition or not.

        Returns
        -------
        Tuple[ArrayFloat1D, ArrayFloat1D]
            left and right step sizes of each interval.
        """

        if len(max_dl_list) != len(len_interval_list):
            raise SetupError(
                "Maximal step size list should have the same length as len_interval_list."
            )

        # left and right step sizes based on maximal step size list
        right_dl = np.roll(max_dl_list, shift=-1)
        left_dl = np.roll(max_dl_list, shift=1)
        # consideration for the first and last interval
        if not is_periodic:
            right_dl[-1] = max_dl_list[-1]
            left_dl[0] = max_dl_list[0]

        # Right and left step size that will be applied for each interval
        right_dl = np.minimum(max_dl_list, right_dl)
        left_dl = np.minimum(max_dl_list, left_dl)

        # Update left and right neighbor step size considering the impact of neighbor intervals
        refine_analy = 1

        while refine_analy > 0:
            refine_analy = 0
            # from left to right, grow to fill up len_interval, minimal 1 step
            tmp_step = 1 - len_interval_list / left_dl * (1 - max_scale)
            num_step = np.maximum(np.log(tmp_step) / np.log(max_scale), 1)
            left_to_right_dl = left_dl * max_scale ** (num_step - 1)
            update_ind = left_to_right_dl < right_dl
            right_dl[update_ind] = left_to_right_dl[update_ind]

            if not is_periodic:
                update_ind[-1] = False

            if np.any(update_ind):
                refine_analy = 1
                left_dl[np.roll(update_ind, shift=1)] = left_to_right_dl[update_ind]

            # from right to left, grow to fill up len_interval, minimal 1 step
            tmp_step = 1 - len_interval_list / right_dl * (1 - max_scale)
            num_step = np.maximum(np.log(tmp_step) / np.log(max_scale), 1)
            right_to_left_dl = right_dl * max_scale ** (num_step - 1)
            update_ind = right_to_left_dl < left_dl
            left_dl[update_ind] = right_to_left_dl[update_ind]

            if not is_periodic:
                update_ind[0] = False

            if np.any(update_ind):
                refine_analy = 1
                right_dl[np.roll(update_ind, shift=-1)] = right_to_left_dl[update_ind]

        if not is_periodic:
            left_dl[0] = max_dl_list[0]
            right_dl[-1] = max_dl_list[-1]

        return left_dl, right_dl

    def make_grid_in_interval(
        self,
        left_neighbor_dl: float,
        right_neighbor_dl: float,
        max_dl: float,
        max_scale: float,
        len_interval: float,
    ) -> ArrayFloat1D:
        """Create a set of grid steps in an interval of length ``len_interval``,
        with first step no larger than ``max_scale * left_neighbor_dl`` and last step no larger than
        ``max_scale * right_neighbor_dl``, with maximum ratio ``max_scale`` between
        neighboring steps. All steps should be no larger than ``max_dl``.

        Parameters
        ----------
        left_neighbor_dl : float
            Step size to left boundary of the interval.
        right_neighbor_dl : float
            Step size to right boundary of the interval.
        max_dl : float
            Maximal step size within the interval.
        max_scale : float
            Maximal ratio between consecutive steps.
        len_interval: float
            Length of the interval.

        Returns
        -------
        ArrayFloat1D
            A list of step sizes in the interval.
        """

        # some validations
        if left_neighbor_dl <= 0 or right_neighbor_dl <= 0 or max_dl <= 0:
            raise ValidationError("Step size needs to be positive.")
        if len_interval <= 0:
            raise ValidationError("The length of the interval must be larger than 0.")
        if max_scale < 1:
            raise ValidationError("max_scale cannot be smaller than 1.")

        # first and last step size
        left_dl = min(max_dl, left_neighbor_dl)
        right_dl = min(max_dl, right_neighbor_dl)

        # classifications:
        grid_type = self.grid_type_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)

        # single pixel
        if grid_type == -1:
            return np.array([len_interval])

        # uniform and multiple pixels
        if grid_type == 0:
            even_dl = min(left_dl, right_dl)
            num_cells = int(np.ceil(len_interval / even_dl))
            return np.array([len_interval / num_cells] * num_cells)

        # grid_type = 1
        # We first set up grid steps from small to large, and then flip
        # their order if right_dl < left_dl
        small_dl = min(left_dl, right_dl)
        large_dl = max(left_dl, right_dl)
        if grid_type == 1:
            # Can small_dl scale to large_dl under max_scale within interval?
            # Compute the number of steps it takes to scale from small_dl to large_dl
            # Check the remaining length in the interval
            num_step = 1 + int(np.floor(np.log(large_dl / small_dl) / np.log(max_scale)))
            len_scale = small_dl * (1 - max_scale**num_step) / (1 - max_scale)
            len_remaining = len_interval - len_scale

            # 1) interval length too small, cannot increase to large_dl, or barely can,
            #    but the remaining part is less than large_dl
            if len_remaining < large_dl:
                dl_list = self.grid_grow_in_interval(small_dl, max_scale, len_interval)
                return dl_list if left_dl <= right_dl else np.flip(dl_list)

            # 2) interval length sufficient, so it will plateau towards large_dl
            dl_list = self.grid_grow_plateau_in_interval(
                small_dl, large_dl, max_scale, len_interval
            )
            return dl_list if left_dl <= right_dl else np.flip(dl_list)

        # grid_type = 2
        if grid_type == 2:
            # Will it be able to plateau?
            # Compute the number of steps it take for both sides to grow to max_it;
            # then compare the length to len_interval
            num_left_step = 1 + int(np.floor(np.log(max_dl / left_dl) / np.log(max_scale)))
            num_right_step = 1 + int(np.floor(np.log(max_dl / right_dl) / np.log(max_scale)))
            len_left = left_dl * (1 - max_scale**num_left_step) / (1 - max_scale)
            len_right = right_dl * (1 - max_scale**num_right_step) / (1 - max_scale)

            len_remaining = len_interval - len_left - len_right

            # able to plateau
            if len_remaining >= max_dl:
                return self.grid_grow_plateau_decrease_in_interval(
                    left_dl, right_dl, max_dl, max_scale, len_interval
                )

            # unable to plateau
            return self.grid_grow_decrease_in_interval(left_dl, right_dl, max_scale, len_interval)

        # unlikely to reach here. For future implementation purpose.
        raise ValidationError("Unimplemented grid type.")

    @staticmethod
    def grid_grow_plateau_decrease_in_interval(
        left_dl: float,
        right_dl: float,
        max_dl: float,
        max_scale: float,
        len_interval: float,
    ) -> ArrayFloat1D:
        """In an interval, grid grows, plateau, and decrease, resembling Lambda letter but
        with plateau in the connection part..

        Parameters
        ----------
        left_dl : float
            Step size at the left boundary.
        right_dl : float
            Step size at the right boundary.
        max_dl : float
            Maximal step size within the interval.
        max_scale : float
            Maximal ratio between consecutive steps.
        len_interval : float
            Length of the interval.
        """

        # Maximum number of steps for undershooting max_dl
        num_left_step = 1 + int(np.floor(np.log(max_dl / left_dl) / np.log(max_scale)))
        num_right_step = 1 + int(np.floor(np.log(max_dl / right_dl) / np.log(max_scale)))

        # step list, in ascending order
        dl_list_left = np.array([left_dl * max_scale**i for i in range(num_left_step)])
        dl_list_right = np.array([right_dl * max_scale**i for i in range(num_right_step)])

        # length
        len_left = left_dl * (1 - max_scale**num_left_step) / (1 - max_scale)
        len_right = right_dl * (1 - max_scale**num_right_step) / (1 - max_scale)

        # remaining part for constant large_dl
        num_const_step = int(np.floor((len_interval - len_left - len_right) / max_dl))
        dl_list_const = np.array([max_dl] * num_const_step)
        len_const = num_const_step * max_dl

        # mismatch
        len_mismatch = len_interval - len_left - len_right - len_const

        # (1) happens to be the right length
        if isclose(len_mismatch, 0):
            return np.concatenate((dl_list_left, dl_list_const, np.flip(dl_list_right)))

        # (2) sufficient remaining part, can be inserted to left or right
        if len_mismatch >= left_dl:
            index_mis = np.searchsorted(dl_list_left, len_mismatch)
            dl_list_left = np.insert(dl_list_left, index_mis, len_mismatch)
            return np.concatenate((dl_list_left, dl_list_const, np.flip(dl_list_right)))

        if len_mismatch >= right_dl:
            index_mis = np.searchsorted(dl_list_right, len_mismatch)
            dl_list_right = np.insert(dl_list_right, index_mis, len_mismatch)
            return np.concatenate((dl_list_left, dl_list_const, np.flip(dl_list_right)))

        # nothing more we can do, let's just add smallest step size,
        # and scale each pixel
        if left_dl <= right_dl:
            dl_list_left = np.append(left_dl, dl_list_left)
        else:
            dl_list_right = np.append(right_dl, dl_list_right)
        dl_list = np.concatenate((dl_list_left, dl_list_const, np.flip(dl_list_right)))
        dl_list *= len_interval / np.sum(dl_list)
        return dl_list

    @staticmethod
    def grid_grow_decrease_in_interval(
        left_dl: float,
        right_dl: float,
        max_scale: float,
        len_interval: float,
    ) -> ArrayFloat1D:
        """In an interval, grid grows, and decrease, resembling Lambda letter.

        Parameters
        ----------
        left_dl : float
            Step size at the left boundary.
        right_dl : float
            Step size at the right boundary.
        max_scale : float
            Maximal ratio between consecutive steps.
        len_interval : float
            Length of the interval.
        """

        # interval too small, it shouldn't happen if bounding box filter is properly handled
        # just use uniform griding with min(left_dl, right_dl)
        if len_interval < left_dl + right_dl:
            even_dl = min(left_dl, right_dl)
            num_cells = int(np.floor(len_interval / even_dl))
            size_snapped = num_cells * even_dl
            if size_snapped < len_interval:
                num_cells += 1
            return np.array([len_interval / num_cells] * num_cells)

        # The maximal number of steps for both sides to undershoot the interval,
        # assuming the last step size from both sides grow to the same size before
        # taking ``floor`` to take integar number of steps.

        # The advantage is that even after taking integar number of steps cutoff,
        # the last step size from the two side will not viloate max_scale.

        tmp_num_l = ((left_dl + right_dl) - len_interval * (1 - max_scale)) / 2 / left_dl
        tmp_num_r = ((left_dl + right_dl) - len_interval * (1 - max_scale)) / 2 / right_dl
        num_left_step = max(int(np.floor(np.log(tmp_num_l) / np.log(max_scale))), 0)
        num_right_step = max(int(np.floor(np.log(tmp_num_r) / np.log(max_scale))), 0)

        # step list, in ascending order
        dl_list_left = np.array([left_dl * max_scale**i for i in range(num_left_step)])
        dl_list_right = np.array([right_dl * max_scale**i for i in range(num_right_step)])

        # length
        len_left = left_dl * (1 - max_scale**num_left_step) / (1 - max_scale)
        len_right = right_dl * (1 - max_scale**num_right_step) / (1 - max_scale)

        # mismatch
        len_mismatch = len_interval - len_left - len_right

        # (1) happens to be the right length
        if isclose(len_mismatch, 0):
            return np.append(dl_list_left, np.flip(dl_list_right))

        # if len_mismatch is larger than the last step size, insert the last step
        while len(dl_list_left) > 0 and len_mismatch >= dl_list_left[-1]:
            dl_list_left = np.append(dl_list_left, dl_list_left[-1])
            len_mismatch -= dl_list_left[-1]

        while len(dl_list_right) > 0 and len_mismatch >= dl_list_right[-1]:
            dl_list_right = np.append(dl_list_right, dl_list_right[-1])
            len_mismatch -= dl_list_right[-1]

        # (2) sufficient remaining part, can be inserted to dl_left or right
        if len_mismatch >= left_dl:
            index_mis = np.searchsorted(dl_list_left, len_mismatch)
            dl_list_left = np.insert(dl_list_left, index_mis, len_mismatch)
            return np.append(dl_list_left, np.flip(dl_list_right))

        if len_mismatch >= right_dl:
            index_mis = np.searchsorted(dl_list_right, len_mismatch)
            dl_list_right = np.insert(dl_list_right, index_mis, len_mismatch)
            return np.append(dl_list_left, np.flip(dl_list_right))

        # nothing more we can do, let's just add smallest step size,
        # and scale each pixel
        if left_dl <= right_dl:
            dl_list_left = np.append(left_dl, dl_list_left)
        else:
            dl_list_right = np.append(right_dl, dl_list_right)
        dl_list = np.append(dl_list_left, np.flip(dl_list_right))
        dl_list *= len_interval / np.sum(dl_list)
        return dl_list

    @staticmethod
    def grid_grow_plateau_in_interval(
        small_dl: float,
        large_dl: float,
        max_scale: float,
        len_interval: float,
    ) -> ArrayFloat1D:
        """In an interval, grid grows, then plateau.

        Parameters
        ----------
        small_dl : float
            The smaller one of step size at the left and right boundaries.
        large_dl : float
            The larger one of step size at the left and right boundaries.
        max_scale : float
            Maximal ratio between consecutive steps.
        len_interval : float
            Length of the interval.

        Returns
        -------
        ArrayFloat1D
            A list of step sizes in the interval, in ascending order.
        """
        # steps for scaling
        num_scale_step = 1 + int(np.floor(np.log(large_dl / small_dl) / np.log(max_scale)))
        dl_list_scale = np.array([small_dl * max_scale**i for i in range(num_scale_step)])
        len_scale = small_dl * (1 - max_scale**num_scale_step) / (1 - max_scale)

        # remaining part for constant large_dl
        num_const_step = int(np.floor((len_interval - len_scale) / large_dl))
        dl_list_const = np.array([large_dl] * num_const_step)
        len_const = large_dl * num_const_step

        # mismatch
        len_mismatch = len_interval - len_scale - len_const

        # (1) happens to be the right length
        if isclose(len_mismatch, 0):
            return np.append(dl_list_scale, dl_list_const)

        # (2) sufficient remaining part, can be inserted to dl_list_scale
        if len_mismatch >= small_dl:
            index_mis = np.searchsorted(dl_list_scale, len_mismatch)
            dl_list_scale = np.insert(dl_list_scale, index_mis, len_mismatch)
            return np.append(dl_list_scale, dl_list_const)

        # nothing more we can do, let's just add smallest step size,
        # and scale each pixel
        dl_list_scale = np.append(small_dl, dl_list_scale)
        dl_list = np.append(dl_list_scale, dl_list_const)
        dl_list *= len_interval / np.sum(dl_list)
        return dl_list

    @staticmethod
    def grid_grow_in_interval(
        small_dl: float,
        max_scale: float,
        len_interval: float,
    ) -> ArrayFloat1D:
        """Mesh simply grows in an interval.

        Parameters
        ----------
        small_dl : float
            The smaller one of step size at the left and right boundaries.
        max_scale : float
            Maximal ratio between consecutive steps.
        len_interval : float
            Length of the interval.

        Returns
        -------
        ArrayFloat1D
            A list of step sizes in the interval, in ascending order.
        """

        # Maximal number of steps for undershooting the interval.
        tmp_step = 1 - len_interval / small_dl * (1 - max_scale)
        num_step = int(np.floor(np.log(tmp_step) / np.log(max_scale)))

        # assuming num_step grids and scaling = max_scale
        dl_list = np.array([small_dl * max_scale**i for i in range(num_step)])
        size_snapped = small_dl * (1 - max_scale**num_step) / (1 - max_scale)

        # mismatch
        len_mismatch = len_interval - size_snapped

        # (1) happens to be the right length
        if isclose(len_mismatch, 0):
            return dl_list

        # (2) sufficient remaining part, can be inserted
        if len_mismatch >= small_dl:
            index_mis = np.searchsorted(dl_list, len_mismatch)
            dl_list = np.insert(dl_list, index_mis, len_mismatch)
            return dl_list

        # (3) remaining part not sufficient to insert, but will not
        # violate max_scale by repearting 1st step, and the last step to include
        # the mismatch part
        if num_step >= 2 and len_mismatch >= small_dl - (1 - 1.0 / max_scale**2) * dl_list[-1]:
            dl_list = np.append(small_dl, dl_list)
            dl_list[-1] += len_mismatch - small_dl
            return dl_list

        # (4) let's see if we can squeeze something out of smaller scaling.
        # For this case, duplicate the 1st step size.
        len_mismatch_even = len_interval - num_step * small_dl
        if isclose(len_mismatch_even, small_dl):
            return np.array([small_dl] * (num_step + 1))

        if len_mismatch_even > small_dl:

            def fun_scale(new_scale):
                if isclose(new_scale, 1.0):
                    return len_interval - small_dl * (1 + num_step)
                return (
                    len_interval - small_dl * (1 - new_scale**num_step) / (1 - new_scale) - small_dl
                )

            # solve for new scaling factor
            # let's not raise exception here, but manually check the convergence.
            root_scalar = Brentq(raise_on_fail=False, epsilon=_ROOTS_TOL)
            sol_scale = root_scalar(fun_scale, 1, max_scale)

            # convergence check based on pyroots API and manual evaluation of the function.
            if sol_scale.converged and abs(fun_scale(sol_scale.x0)) <= _ROOTS_TOL:
                new_scale = sol_scale.x0
                dl_list = np.array([small_dl * new_scale**i for i in range(num_step)])
                dl_list = np.append(small_dl, dl_list)
                return dl_list
            # if not converged, let's use the strategy below.

        # nothing more we can do, let's just add smallest step size,
        # and scale each pixel
        dl_list = np.append(small_dl, dl_list)
        dl_list *= len_interval / np.sum(dl_list)
        return dl_list

    @staticmethod
    def grid_type_in_interval(
        left_dl: float,
        right_dl: float,
        max_dl: float,
        max_scale: float,
        len_interval: float,
    ) -> int:
        """Mesh type check (in an interval).

        Parameters
        ----------
        left_dl : float
            Step size at left boundary of the interval.
        right_dl : float
            Step size at right boundary of the interval.
        max_dl : float
            Maximal step size within the interval.
        max_scale : float
            Maximal ratio between consecutive steps.
        len_interval: float
            Length of the interval.

        Returns
        -------
        grid_type : int
            -1 for single pixel grid
            0 for uniform grid
            1 for small to large to optionally plateau grid
            2 for small to large to optionally plateau to small grid
        """

        # uniform grid if interval length is no larger than small_dl
        if len_interval <= min(left_dl, right_dl, max_dl):
            return -1
        # uniform grid if max_scale is too small
        if isclose(max_scale, 1):
            return 0
        # uniform grid if max_dl is the smallest
        if max_dl <= left_dl and max_dl <= right_dl:
            return 0

        # type 1
        if max_dl <= left_dl or max_dl <= right_dl:
            return 1

        return 2


MesherType = Union[GradedMesher]



"""Monitor level data, store the DataArrays associated with a single heat-charge monitor."""

from __future__ import annotations

from typing import Union

from tidy3d.components.tcad.data.monitor_data.charge import (
    SteadyCapacitanceData,
    SteadyFreeCarrierData,
    SteadyPotentialData,
)
from tidy3d.components.tcad.data.monitor_data.heat import TemperatureData

TCADMonitorDataType = Union[
    TemperatureData,
    SteadyPotentialData,
    SteadyFreeCarrierData,
    SteadyCapacitanceData,
]



from .base import Expression
from .functions import Cos, Exp, Log, Log10, Sin, Sqrt, Tan
from .metrics import ModeAmp, ModePower, generate_validation_data
from .variables import Constant, Variable

__all__ = [
    "Expression",
    "Constant",
    "Variable",
    "ModeAmp",
    "ModePower",
    "generate_validation_data",
    "Sin",
    "Cos",
    "Tan",
    "Exp",
    "Log",
    "Log10",
    "Sqrt",
]

# The following code dynamically collects all classes that are subclasses of Expression
# from the specified modules and updates their forward references. This is necessary to handle
# cases where classes reference each other before they are fully defined. The local_vars dictionary
# is used to store these classes and any other necessary types for the forward reference updates.

import importlib
import inspect

from .types import ExpressionType

_module_names = ["base", "variables", "functions", "metrics", "operators"]
_model_classes = set()
_local_vars = {"ExpressionType": ExpressionType}

for module_name in _module_names:
    module = importlib.import_module(f".{module_name}", package=__name__)
    for name, obj in inspect.getmembers(module):
        if inspect.isclass(obj) and issubclass(obj, Expression):
            _model_classes.add(obj)
            _local_vars[name] = obj

for cls in _model_classes:
    cls.update_forward_refs(**_local_vars)



"""Mesh-defined geometry."""

from __future__ import annotations

from abc import ABC
from typing import List, Optional, Tuple, Union

import numpy as np
import pydantic.v1 as pydantic

from ...constants import inf
from ...exceptions import DataError, ValidationError
from ...log import log
from ...packaging import verify_packages_import
from ..base import cached_property
from ..data.data_array import DATA_ARRAY_MAP, TriangleMeshDataArray
from ..data.dataset import TriangleMeshDataset
from ..data.validators import validate_no_nans
from ..types import Ax, Bound, Coordinate, MatrixReal4x4, Shapely
from ..viz import add_ax_if_none, equal_aspect
from . import base

AREA_SIZE_THRESHOLD = 1e-36


class TriangleMesh(base.Geometry, ABC):
    """Custom surface geometry given by a triangle mesh, as in the STL file format.

    Example
    -------
    >>> vertices = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]])
    >>> faces = np.array([[1, 2, 3], [0, 3, 2], [0, 1, 3], [0, 2, 1]])
    >>> stl_geom = TriangleMesh.from_vertices_faces(vertices, faces)
    """

    mesh_dataset: Optional[TriangleMeshDataset] = pydantic.Field(
        ...,
        title="Surface mesh data",
        description="Surface mesh data.",
    )

    _no_nans_mesh = validate_no_nans("mesh_dataset")

    @pydantic.root_validator(pre=True)
    @verify_packages_import(["trimesh"])
    def _validate_trimesh_library(cls, values):
        """Check if the trimesh package is imported as a validator."""
        return values

    @pydantic.validator("mesh_dataset", pre=True, always=True)
    def _warn_if_none(cls, val: TriangleMeshDataset) -> TriangleMeshDataset:
        """Warn if the Dataset fails to load."""
        if isinstance(val, dict):
            if any((v in DATA_ARRAY_MAP for _, v in val.items() if isinstance(v, str))):
                log.warning("Loading 'mesh_dataset' without data.")
                return None
        return val

    @pydantic.validator("mesh_dataset", always=True)
    @verify_packages_import(["trimesh"])
    def _check_mesh(cls, val: TriangleMeshDataset) -> TriangleMeshDataset:
        """Check that the mesh is valid."""
        if val is None:
            return None

        import trimesh

        mesh = cls._triangles_to_trimesh(val.surface_mesh)
        if not all(np.array(mesh.area_faces) > AREA_SIZE_THRESHOLD):
            old_tol = trimesh.tol.merge
            trimesh.tol.merge = np.sqrt(2 * AREA_SIZE_THRESHOLD)
            new_mesh = mesh.process(validate=True)
            trimesh.tol.merge = old_tol
            val = TriangleMesh.from_trimesh(new_mesh).mesh_dataset
            log.warning(
                f"The provided mesh has triangles with near zero area < {AREA_SIZE_THRESHOLD}. "
                "Triangles which have one edge of their 2D oriented bounding box shorter than "
                f"'sqrt(2*{AREA_SIZE_THRESHOLD}) are being automatically removed.'"
            )
            if not all(np.array(new_mesh.area_faces) > AREA_SIZE_THRESHOLD):
                raise ValidationError(
                    f"The provided mesh has triangles with near zero area < {AREA_SIZE_THRESHOLD}. "
                    "The automatic removal of these triangles has failed. You can try "
                    "using numpy-stl's 'from_file' import with 'remove_empty_areas' set "
                    "to True and a suitable 'AREA_SIZE_THRESHOLD' to remove them."
                )
        if not mesh.is_watertight:
            log.warning(
                "The provided mesh is not watertight. "
                "This can lead to incorrect permittivity distributions, "
                "and can also cause problems with plotting and mesh validation. "
                "You can try 'TriangleMesh.fill_holes', which attempts to repair the mesh. "
                "Otherwise, the mesh may require manual repair. You can use a "
                "'PermittivityMonitor' to check if the permittivity distribution is correct. "
                "You can see which faces are broken using 'trimesh.repair.broken_faces'."
            )
        if not mesh.is_winding_consistent:
            log.warning(
                "The provided mesh does not have consistent winding (face orientations). "
                "This can lead to incorrect permittivity distributions, "
                "and can also cause problems with plotting and mesh validation. "
                "You can try 'TriangleMesh.fix_winding', which attempts to repair the mesh. "
                "Otherwise, the mesh may require manual repair. You can use a "
                "'PermittivityMonitor' to check if the permittivity distribution is correct. "
            )
        if not mesh.is_volume:
            log.warning(
                "The provided mesh does not represent a valid volume, possibly due to "
                "incorrect normal vector orientation. "
                "This can lead to incorrect permittivity distributions, "
                "and can also cause problems with plotting and mesh validation. "
                "You can try 'TriangleMesh.fix_normals', "
                "which attempts to fix the normals to be consistent and outward-facing. "
                "Otherwise, the mesh may require manual repair. You can use a "
                "'PermittivityMonitor' to check if the permittivity distribution is correct."
            )

        return val

    @verify_packages_import(["trimesh"])
    def fix_winding(self) -> TriangleMesh:
        """Try to fix winding in the mesh."""
        import trimesh

        mesh = TriangleMesh._triangles_to_trimesh(self.mesh_dataset.surface_mesh)
        trimesh.repair.fix_winding(mesh)
        return TriangleMesh.from_trimesh(mesh)

    @verify_packages_import(["trimesh"])
    def fill_holes(self) -> TriangleMesh:
        """Try to fill holes in the mesh. Can be used to repair non-watertight meshes."""
        import trimesh

        mesh = TriangleMesh._triangles_to_trimesh(self.mesh_dataset.surface_mesh)
        trimesh.repair.fill_holes(mesh)
        return TriangleMesh.from_trimesh(mesh)

    @verify_packages_import(["trimesh"])
    def fix_normals(self) -> TriangleMesh:
        """Try to fix normals to be consistent and outward-facing."""
        import trimesh

        mesh = TriangleMesh._triangles_to_trimesh(self.mesh_dataset.surface_mesh)
        trimesh.repair.fix_normals(mesh)
        return TriangleMesh.from_trimesh(mesh)

    @classmethod
    @verify_packages_import(["trimesh"])
    def from_stl(
        cls,
        filename: str,
        scale: float = 1.0,
        origin: Tuple[float, float, float] = (0, 0, 0),
        solid_index: int = None,
        **kwargs,
    ) -> Union[TriangleMesh, base.GeometryGroup]:
        """Load a :class:`.TriangleMesh` directly from an STL file.
        The ``solid_index`` parameter can be used to select a single solid from the file.
        Otherwise, if the file contains a single solid, it will be loaded as a
        :class:`.TriangleMesh`; if the file contains multiple solids,
        they will all be loaded as a :class:`.GeometryGroup`.

        Parameters
        ----------
        filename : str
            The name of the STL file containing the surface geometry mesh data.
        scale : float = 1.0
            The length scale for the loaded geometry (um).
            For example, a scale of 10.0 means that a vertex (1, 0, 0) will be placed at
            x = 10 um.
        origin : Tuple[float, float, float] = (0, 0, 0)
            The origin of the loaded geometry, in units of ``scale``.
            Translates from (0, 0, 0) to this point after applying the scaling.
        solid_index : int = None
            If set, read a single solid with this index from the file.

        Returns
        -------
        Union[:class:`.TriangleMesh`, :class:`.GeometryGroup`]
            The geometry or geometry group from the file.
        """
        import trimesh

        from ..types_extra import TrimeshType

        def process_single(mesh: TrimeshType) -> TriangleMesh:
            """Process a single 'trimesh.Trimesh' using scale and origin."""
            mesh.apply_scale(scale)
            mesh.apply_translation(origin)
            return cls.from_trimesh(mesh)

        scene = trimesh.load(filename, **kwargs)
        meshes = []
        if isinstance(scene, trimesh.Trimesh):
            meshes = [scene]
        elif isinstance(scene, trimesh.Scene):
            meshes = scene.dump()
        else:
            raise ValidationError(
                "Invalid trimesh type in file. Supported types are 'trimesh.Trimesh' "
                "and 'trimesh.Scene'."
            )

        if solid_index is None:
            if isinstance(scene, trimesh.Trimesh):
                return process_single(scene)
            if isinstance(scene, trimesh.Scene):
                geoms = [process_single(mesh) for mesh in meshes]
                return base.GeometryGroup(geometries=geoms)

        if solid_index < len(meshes):
            return process_single(meshes[solid_index])
        raise ValidationError("No solid found at 'solid_index' in the stl file.")

    @classmethod
    @verify_packages_import(["trimesh"])
    def from_trimesh(cls, mesh: trimesh.Trimesh) -> TriangleMesh:
        """Create a :class:`.TriangleMesh` from a ``trimesh.Trimesh`` object.

        Parameters
        ----------
        trimesh : ``trimesh.Trimesh``
            The Trimesh object containing the surface geometry mesh data.

        Returns
        -------
        :class:`.TriangleMesh`
            The custom surface mesh geometry given by the ``trimesh.Trimesh`` provided.
        """
        return cls.from_vertices_faces(mesh.vertices, mesh.faces)

    @classmethod
    def from_triangles(cls, triangles: np.ndarray) -> TriangleMesh:
        """Create a :class:`.TriangleMesh` from a numpy array
        containing the triangles of a surface mesh.

        Parameters
        ----------
        triangles : ``np.ndarray``
            A numpy array of shape (N, 3, 3) storing the triangles of the surface mesh.
            The first index labels the triangle, the second index labels the vertex
            within a given triangle, and the third index is the coordinate (x, y, or z).

        Returns
        -------
        :class:`.TriangleMesh`
            The custom surface mesh geometry given by the triangles provided.

        """
        triangles = np.array(triangles)
        if len(triangles.shape) != 3 or triangles.shape[1] != 3 or triangles.shape[2] != 3:
            raise ValidationError(
                f"Provided 'triangles' must be an N x 3 x 3 array, given {triangles.shape}."
            )
        num_faces = len(triangles)
        coords = dict(
            face_index=np.arange(num_faces),
            vertex_index=np.arange(3),
            axis=np.arange(3),
        )
        vertices = TriangleMeshDataArray(triangles, coords=coords)
        mesh_dataset = TriangleMeshDataset(surface_mesh=vertices)
        return TriangleMesh(mesh_dataset=mesh_dataset)

    @classmethod
    @verify_packages_import(["trimesh"])
    def from_vertices_faces(cls, vertices: np.ndarray, faces: np.ndarray) -> TriangleMesh:
        """Create a :class:`.TriangleMesh` from numpy arrays containing the data
        of a surface mesh. The first array contains the vertices, and the second array contains
        faces formed from triples of the vertices.

        Parameters
        ----------
        vertices: ``np.ndarray``
            A numpy array of shape (N, 3) storing the vertices of the surface mesh.
            The first index labels the vertex, and the second index is the coordinate
            (x, y, or z).
        faces : ``np.ndarray``
            A numpy array of shape (M, 3) storing the indices of the vertices of each face
            in the surface mesh. The first index labels the face, and the second index
            labels the vertex index within the ``vertices`` array.

        Returns
        -------
        :class:`.TriangleMesh`
            The custom surface mesh geometry given by the vertices and faces provided.

        """
        import trimesh

        vertices = np.array(vertices)
        faces = np.array(faces)
        if len(vertices.shape) != 2 or vertices.shape[1] != 3:
            raise ValidationError(
                f"Provided 'vertices' must be an N x 3 array, given {vertices.shape}."
            )
        if len(faces.shape) != 2 or faces.shape[1] != 3:
            raise ValidationError(f"Provided 'faces' must be an M x 3 array, given {faces.shape}.")
        return cls.from_triangles(trimesh.Trimesh(vertices, faces).triangles)

    @classmethod
    @verify_packages_import(["trimesh"])
    def _triangles_to_trimesh(
        cls, triangles: np.ndarray
    ):  # -> TrimeshType: We need to get this out of the classes and into functional methods operating on a class (maybe still referenced to the class)
        """Convert an (N, 3, 3) numpy array of triangles to a ``trimesh.Trimesh``."""
        import trimesh

        return trimesh.Trimesh(**trimesh.triangles.to_kwargs(triangles))

    @cached_property
    @verify_packages_import(["trimesh"])
    def trimesh(
        self,
    ):  # -> TrimeshType: We need to get this out of the classes and into functional methods operating on a class (maybe still referenced to the class)
        """A ``trimesh.Trimesh`` object representing the custom surface mesh geometry."""
        return self._triangles_to_trimesh(self.triangles)

    @cached_property
    def triangles(self) -> np.ndarray:
        """The triangles of the surface mesh as an ``np.ndarray``."""
        if self.mesh_dataset is None:
            raise DataError("Can't get triangles as 'mesh_dataset' is None.")
        return self.mesh_dataset.surface_mesh.to_numpy()

    def _surface_area(self, bounds: Bound) -> float:
        """Returns object's surface area within given bounds."""
        # currently ignores bounds
        return self.trimesh.area

    def _volume(self, bounds: Bound) -> float:
        """Returns object's volume within given bounds."""
        # currently ignores bounds
        return self.trimesh.volume

    @cached_property
    def bounds(self) -> Bound:
        """Returns bounding box min and max coordinates.

        Returns
        -------
        Tuple[float, float, float], Tuple[float, float float]
            Min and max bounds packaged as ``(minx, miny, minz), (maxx, maxy, maxz)``.
        """
        if self.mesh_dataset is None:
            return ((-inf, -inf, -inf), (inf, inf, inf))
        return self.trimesh.bounds

    def intersections_tilted_plane(
        self, normal: Coordinate, origin: Coordinate, to_2D: MatrixReal4x4
    ) -> List[Shapely]:
        """Return a list of shapely geometries at the plane specified by normal and origin.

        Parameters
        ----------
        normal : Coordinate
            Vector defining the normal direction to the plane.
        origin : Coordinate
            Vector defining the plane origin.
        to_2D : MatrixReal4x4
            Transformation matrix to apply to resulting shapes.

        Returns
        -------
        List[shapely.geometry.base.BaseGeometry]
            List of 2D shapes that intersect plane.
            For more details refer to
            `Shapely's Documentation <https://shapely.readthedocs.io/en/stable/project.html>`_.
        """
        section = self.trimesh.section(plane_origin=origin, plane_normal=normal)
        if section is None:
            return []
        path, _ = section.to_planar(to_2D=to_2D)
        return path.polygons_full

    def intersections_plane(
        self, x: float = None, y: float = None, z: float = None
    ) -> List[Shapely]:
        """Returns list of shapely geometries at plane specified by one non-None value of x,y,z.

        Parameters
        ----------
        x : float = None
            Position of plane in x direction, only one of x,y,z can be specified to define plane.
        y : float = None
            Position of plane in y direction, only one of x,y,z can be specified to define plane.
        z : float = None
            Position of plane in z direction, only one of x,y,z can be specified to define plane.

        Returns
        -------
        List[shapely.geometry.base.BaseGeometry]
            List of 2D shapes that intersect plane.
            For more details refer to
            `Shapely's Documentaton <https://shapely.readthedocs.io/en/stable/project.html>`_.
        """

        if self.mesh_dataset is None:
            return []

        axis, position = self.parse_xyz_kwargs(x=x, y=y, z=z)

        origin = self.unpop_axis(position, (0, 0), axis=axis)
        normal = self.unpop_axis(1, (0, 0), axis=axis)

        mesh = self.trimesh

        try:
            section = mesh.section(plane_origin=origin, plane_normal=normal)

            if section is None:
                return []

            # homogeneous transformation matrix to map to xy plane
            mapping = np.eye(4)

            # translate to origin
            mapping[3, :3] = -np.array(origin)

            # permute so normal is aligned with z axis
            # and (y, z), (x, z), resp. (x, y) are aligned with (x, y)
            identity = np.eye(3)
            permutation = self.unpop_axis(identity[2], identity[0:2], axis=axis)
            mapping[:3, :3] = np.array(permutation).T

            section2d, _ = section.to_planar(to_2D=mapping)
            return list(section2d.polygons_full)

        except ValueError as e:
            if not mesh.is_watertight:
                log.warning(
                    "Unable to compute 'TriangleMesh.intersections_plane' "
                    "because the mesh was not watertight. Using bounding box instead. "
                    "This may be overly strict; consider using 'TriangleMesh.fill_holes' "
                    "to repair the non-watertight mesh."
                )
            else:
                log.warning(
                    "Unable to compute 'TriangleMesh.intersections_plane'. "
                    "Using bounding box instead."
                )
            log.warning(f"Error encountered: {e}")
            return self.bounding_box.intersections_plane(x=x, y=y, z=z)

    def inside(
        self, x: np.ndarray[float], y: np.ndarray[float], z: np.ndarray[float]
    ) -> np.ndarray[bool]:
        """For input arrays ``x``, ``y``, ``z`` of arbitrary but identical shape, return an array
        with the same shape which is ``True`` for every point in zip(x, y, z) that is inside the
        volume of the :class:`Geometry`, and ``False`` otherwise.

        Parameters
        ----------
        x : np.ndarray[float]
            Array of point positions in x direction.
        y : np.ndarray[float]
            Array of point positions in y direction.
        z : np.ndarray[float]
            Array of point positions in z direction.

        Returns
        -------
        np.ndarray[bool]
            ``True`` for every point that is inside the geometry.
        """

        arrays = tuple(map(np.array, (x, y, z)))
        self._ensure_equal_shape(*arrays)
        inside = np.zeros((arrays[0].size,), dtype=bool)
        arrays_flat = map(np.ravel, arrays)
        arrays_stacked = np.stack(tuple(arrays_flat), axis=-1)
        inside = self.trimesh.contains(arrays_stacked)
        return inside.reshape(arrays[0].shape)

    @equal_aspect
    @add_ax_if_none
    def plot(
        self, x: float = None, y: float = None, z: float = None, ax: Ax = None, **patch_kwargs
    ) -> Ax:
        """Plot geometry cross section at single (x,y,z) coordinate.

        Parameters
        ----------
        x : float = None
            Position of plane in x direction, only one of x,y,z can be specified to define plane.
        y : float = None
            Position of plane in y direction, only one of x,y,z can be specified to define plane.
        z : float = None
            Position of plane in z direction, only one of x,y,z can be specified to define plane.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        **patch_kwargs
            Optional keyword arguments passed to the matplotlib patch plotting of structure.
            For details on accepted values, refer to
            `Matplotlib's documentation <https://tinyurl.com/2nf5c2fk>`_.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """

        log.warning(
            "Plotting a 'TriangleMesh' may give inconsistent results "
            "if the mesh is not unionized. We recommend unionizing all meshes before import. "
            "A 'PermittivityMonitor' can be used to check that the mesh is loaded correctly."
        )

        return base.Geometry.plot(self, x=x, y=y, z=z, ax=ax, **patch_kwargs)



"""Convenience functions for estimating antenna radiation by applying array factor."""

from abc import ABC, abstractmethod
from typing import Optional, Tuple, Union

import numpy as np
import pydantic.v1 as pd
from pydantic.v1 import NonNegativeFloat, PositiveInt

from ...components.base import Tidy3dBaseModel, skip_if_fields_missing
from ...components.data.monitor_data import AbstractFieldProjectionData, DirectivityData
from ...components.data.sim_data import SimulationData
from ...components.geometry.base import Box, Geometry
from ...components.grid.grid_spec import GridSpec, LayerRefinementSpec
from ...components.lumped_element import LumpedElement
from ...components.medium import Medium, MediumType3D
from ...components.monitor import AbstractFieldProjectionMonitor, MonitorType
from ...components.simulation import Simulation
from ...components.source.utils import SourceType
from ...components.structure import MeshOverrideStructure, Structure
from ...components.types import ArrayLike, Axis, Bound
from ...constants import C_0, inf
from ...log import log


class AbstractAntennaArrayCalculator(Tidy3dBaseModel, ABC):
    """Abstract base for phased array calculators."""

    @property
    @abstractmethod
    def _antenna_locations(self) -> ArrayLike:
        """Locations of antennas' centers in an array."""

    @property
    @abstractmethod
    def _antenna_amps(self) -> ArrayLike:
        """Amplitude multipliers of antennas in an array."""

    @property
    @abstractmethod
    def _antenna_phases(self) -> ArrayLike:
        """Phase shifts of antennas in an array."""

    @property
    @abstractmethod
    def _extend_dims(self):
        """Dimensions along which antennas will be duplicated."""

    @property
    def _antenna_nominal_size(self) -> ArrayLike:
        """Size of the antenna array without taking into account size of a single antenna."""
        rmin = np.min(self._antenna_locations, axis=0)
        rmax = np.max(self._antenna_locations, axis=0)

        return rmax - rmin

    @property
    def _antenna_nominal_center(self) -> ArrayLike:
        """Center of the antenna array without taking into account size of a single antenna."""
        rmin = np.min(self._antenna_locations, axis=0)
        rmax = np.max(self._antenna_locations, axis=0)

        return 0.5 * (rmax + rmin)

    def _detect_antenna_bounds(self, simulation: Simulation):
        """Detect the bounds of the antenna in the simulation."""
        # directions in which we will need to tile simulation
        extend_dims = self._extend_dims

        # detect bounding box of all structures, sources, and lumped elements in the simulation
        sim_bounds = np.array(simulation.bounds)
        antenna_bounds = sim_bounds.copy()
        for dim in extend_dims:
            antenna_bounds[0][dim] = inf
            antenna_bounds[1][dim] = -inf

        all_objects = list(simulation.structures)
        all_objects += list(simulation.sources)
        all_objects += list(simulation.lumped_elements)

        for obj in all_objects:
            # get bounding box of the object
            if isinstance(obj, Structure):
                obj_bounds = np.array(obj.geometry.bounds)
            else:
                obj_bounds = np.array(obj.bounds)

            for dim in extend_dims:
                # update minimum and maximum bounds in each dimension
                # check if object extends beyond the simulation bounds on both sides
                extends_beyond_min = obj_bounds[0][dim] < sim_bounds[0][dim]
                extends_beyond_max = obj_bounds[1][dim] > sim_bounds[1][dim]
                if extends_beyond_min or extends_beyond_max:
                    # in case of a box, we just ignore it, since we will be able to extend it later
                    if (
                        isinstance(obj, Structure)
                        and isinstance(obj.geometry, Box)
                        and extends_beyond_min
                        and extends_beyond_max
                    ):
                        continue
                    # otherwise shrink the object bounds to simulation bounds
                    obj_bounds[0][dim] = max(obj_bounds[0][dim], sim_bounds[0][dim])
                    obj_bounds[1][dim] = min(obj_bounds[1][dim], sim_bounds[1][dim])
                    # and show a warning
                    log.warning(
                        f"Object {obj.name} (type: {obj.type}) extends beyond simulation bounds along"
                        f" '{'xyz'[dim]}' axis. Please check your antenna setup for correctness."
                    )
                # update minimum and maximum bounds in each dimension
                antenna_bounds[0][dim] = min(antenna_bounds[0][dim], obj_bounds[0][dim])
                antenna_bounds[1][dim] = max(antenna_bounds[1][dim], obj_bounds[1][dim])

        return antenna_bounds

    def _try_to_expand_geometry(
        self, geometry: Geometry, old_sim_bounds: Bound, new_sim_bounds: Bound
    ):
        """Try to expand geometry to cover the entire simulation domain."""

        can_expand = isinstance(geometry, Box) and all(
            geometry.bounds[0][dim] < old_sim_bounds[0][dim]
            or geometry.bounds[1][dim] > old_sim_bounds[1][dim]
            for dim in self._extend_dims
        )

        if not can_expand:
            return None  # could not expand geometry, so we will duplicate

        # get original structure bounds
        box_bounds = np.array(geometry.bounds)
        box_size = np.array(geometry.size)
        box_center = np.array(geometry.center)

        for dim in self._extend_dims:
            # if it extends beyond simulation bounds, but size is not inf
            # we will adjust it so that it goes out of bounds by the same amount as in the initial simulation.
            # No need to do anything if size is inf.
            if box_size[dim] != np.inf:
                # shift min bounds to new location
                box_bounds[0][dim] = new_sim_bounds[0][dim] + (
                    box_bounds[0][dim] - old_sim_bounds[0][dim]
                )

                # shift max bounds to new location
                box_bounds[1][dim] = new_sim_bounds[1][dim] - (
                    old_sim_bounds[1][dim] - box_bounds[1][dim]
                )

                # calculate new structure size and center
                box_size[dim] = box_bounds[1][dim] - box_bounds[0][dim]
                box_center[dim] = 0.5 * (box_bounds[0][dim] + box_bounds[1][dim])

        return geometry.updated_copy(
            center=tuple(box_center),
            size=tuple(box_size),
        )

    def _duplicate_or_expand_list_of_objects(
        self,
        objects: Tuple[
            Union[Structure, MeshOverrideStructure, LayerRefinementSpec, LumpedElement], ...
        ],
        old_sim_bounds: Bound,
        new_sim_bounds: Bound,
    ):
        """Duplicate or expand a list of objects."""

        locations = self._antenna_locations

        array_objects = []
        for obj in objects:
            if isinstance(obj, (Structure, MeshOverrideStructure)):
                geometry = obj.geometry
                if isinstance(obj, Structure) and obj.medium.is_custom:
                    log.warning(
                        f"Object '{obj.name}' contains a custom medium which has limited support "
                        "in automatic antenna array generation. The custom medium's spatial distribution "
                        "will remain fixed and will not be translated along with the structure geometry."
                    )
            elif isinstance(obj, (LayerRefinementSpec, LumpedElement)):
                geometry = obj
            else:
                raise ValueError(f"Object of type {type(obj)} is not supported.")

            # check if it is a box and extends beyond simulation bounds
            expanded_geometry = self._try_to_expand_geometry(
                geometry=geometry, old_sim_bounds=old_sim_bounds, new_sim_bounds=new_sim_bounds
            )

            if expanded_geometry is None:
                # could not expand geometry, so we duplicate
                for ind, translation_vector in enumerate(locations):
                    if isinstance(obj, Structure):
                        new_obj = obj.updated_copy(
                            geometry=obj.geometry.translated(*translation_vector),
                            name=None if obj.name is None else f"{obj.name}_{ind}",
                        )
                    elif isinstance(obj, MeshOverrideStructure):
                        new_obj = obj.updated_copy(
                            geometry=obj.geometry.translated(*translation_vector).bounding_box,
                            name=None if obj.name is None else f"{obj.name}_{ind}",
                        )
                    elif isinstance(obj, LayerRefinementSpec):
                        new_obj = obj.updated_copy(
                            center=tuple(obj.center + translation_vector),
                        )
                    elif isinstance(obj, LumpedElement):
                        new_obj = obj.updated_copy(
                            center=tuple(obj.center + translation_vector),
                            name=None if obj.name is None else f"{obj.name}_{ind}",
                        )
                    array_objects.append(new_obj)
            else:
                # could expand geometry, so we create a copy of the original object with updated size and center
                if isinstance(obj, (Structure, MeshOverrideStructure)):
                    new_obj = obj.updated_copy(
                        geometry=expanded_geometry,
                    )
                elif isinstance(obj, (LayerRefinementSpec, LumpedElement)):
                    new_obj = expanded_geometry

                # add the new object to the list of objects
                array_objects.append(new_obj)

        return array_objects

    def _expand_monitors(
        self,
        monitors: Tuple[MonitorType, ...],
        antenna_bounds: Bound,
        new_sim_bounds: Bound,
        old_sim_bounds: Bound,
    ):
        """Expand monitors."""

        extend_dims = self._extend_dims

        # expand far-field monitors
        array_monitors = []
        for monitor in monitors:
            # we only expand field projection monitors
            if isinstance(monitor, AbstractFieldProjectionMonitor):
                # get original monitor bounds
                mnt_bounds = np.array(monitor.bounds)
                mnt_size = np.array(monitor.size)
                mnt_center = np.array(monitor.center)

                if any(mnt_size[dim] == 0 for dim in extend_dims):
                    log.warning(
                        f"Monitor '{monitor.name}' (type: '{monitor.type}') has zero size along "
                        f"one of the axes. It will not be included in the resulting simulation."
                    )
                    continue

                for dim in extend_dims:
                    if mnt_size[dim] != np.inf:
                        # check that monitor covers the estimated antenna box
                        if (
                            mnt_bounds[0][dim] > antenna_bounds[0][dim]
                            or mnt_bounds[1][dim] < antenna_bounds[1][dim]
                        ):
                            log.warning(
                                f"Monitor '{monitor.name}' (type: '{monitor.type}') does not cover "
                                f"the estimated antenna box along '{'xyz'[dim]}' axis. "
                                "The automatically extended monitor will likely be wrong. "
                                "Please double check the resulting simulation."
                            )
                        # shift min bounds to new location
                        mnt_bounds[0][dim] = new_sim_bounds[0][dim] + (
                            mnt_bounds[0][dim] - old_sim_bounds[0][dim]
                        )

                        # shift max bounds to new location
                        mnt_bounds[1][dim] = new_sim_bounds[1][dim] - (
                            old_sim_bounds[1][dim] - mnt_bounds[1][dim]
                        )

                        # calculate new monitor size and center
                        mnt_size[dim] = mnt_bounds[1][dim] - mnt_bounds[0][dim]
                        mnt_center[dim] = 0.5 * (mnt_bounds[0][dim] + mnt_bounds[1][dim])

                # create a copy of the original monitor with updated size and center
                new_mnt = monitor.updated_copy(
                    center=tuple(mnt_center),
                    size=tuple(mnt_size),
                )

                # add the new monitor to the list of monitors
                array_monitors.append(new_mnt)

            # otherwise we ignore and warn the user
            else:
                log.warning(
                    f"Monitor '{monitor.name}' (type: '{monitor.type}') will not be automatically "
                    "transferred into the resulting antenna array simulation."
                )

        return array_monitors

    def _duplicate_structures(
        self, structures: Tuple[Structure, ...], new_sim_bounds: Bound, old_sim_bounds: Bound
    ):
        """Duplicate structures."""

        return self._duplicate_or_expand_list_of_objects(
            objects=structures, old_sim_bounds=old_sim_bounds, new_sim_bounds=new_sim_bounds
        )

    def _duplicate_sources(
        self,
        sources: Tuple[SourceType, ...],
        lumped_elements: Tuple[LumpedElement, ...],
        old_sim_bounds: Bound,
        new_sim_bounds: Bound,
    ):
        """Duplicate sources and lumped elements."""
        array_lumped_elements = self._duplicate_or_expand_list_of_objects(
            objects=lumped_elements, old_sim_bounds=old_sim_bounds, new_sim_bounds=new_sim_bounds
        )

        array_sources = []
        for ind, (translation_vector, phase_shift, amp_multiplier) in enumerate(
            zip(self._antenna_locations, self._antenna_phases, self._antenna_amps)
        ):
            if amp_multiplier != 0.0:
                for source in sources:
                    new_source = source.updated_copy(
                        center=tuple(source.center + translation_vector),
                        name=f"{source.name}_{ind}",
                        source_time=source.source_time.updated_copy(
                            phase=source.source_time.phase + phase_shift,
                            amplitude=source.source_time.amplitude * amp_multiplier,
                        ),
                    )
                    array_sources.append(new_source)

        return array_sources, array_lumped_elements

    def _duplicate_grid_specs(
        self, grid_spec: GridSpec, old_sim_bounds: Bound, new_sim_bounds: Bound
    ):
        """Duplicate grid specs."""

        array_overrides = self._duplicate_or_expand_list_of_objects(
            objects=grid_spec.override_structures,
            old_sim_bounds=old_sim_bounds,
            new_sim_bounds=new_sim_bounds,
        )
        array_layer_refinement_specs = self._duplicate_or_expand_list_of_objects(
            objects=grid_spec.layer_refinement_specs,
            old_sim_bounds=old_sim_bounds,
            new_sim_bounds=new_sim_bounds,
        )

        array_snapping_points = []
        for translation_vector in self._antenna_locations:
            for snapping_point in grid_spec.snapping_points:
                new_snapping_point = [
                    snapping_point[dim] + translation_vector[dim]
                    if snapping_point[dim] is not None
                    else None
                    for dim in range(3)
                ]
                array_snapping_points.append(new_snapping_point)

        return grid_spec.updated_copy(
            override_structures=array_overrides,
            layer_refinement_specs=array_layer_refinement_specs,
            snapping_points=array_snapping_points,
        )

    def make_antenna_array(self, simulation: Simulation) -> Simulation:
        """
        Converts a single antenna simulation into an antenna array simulation.
        This function identifies the size and position of the single antenna
        in the input simulation and uses this information to compute the dimensions
        of the resulting antenna array simulation. All structures, sources, lumped elements,
        and mesh override structures are duplicated, while monitors are extended in size.
        Only projection monitors are transferred into the resulting simulation.

        For best results, the antenna assembly should be contained within the simulation bounds.

        Parameters:
        ----------
        simulation : Simulation
            The simulation specification describing a single antenna setup.

        Returns:
        --------
        Simulation
            The simulation specification describing the antenna array.
        """

        # detect bounding box of all structures, sources, and lumped elements in the simulation
        sim_bounds = np.array(simulation.bounds)
        antenna_bounds = self._detect_antenna_bounds(simulation)

        # compute the center and size of the antenna
        antenna_size = antenna_bounds[1] - antenna_bounds[0]
        antenna_center = 0.5 * (antenna_bounds[0] + antenna_bounds[1])

        # compute buffer between the antenna and simulation bounds on each side
        buffer_min = antenna_bounds[0] - sim_bounds[0]
        buffer_max = sim_bounds[1] - antenna_bounds[1]

        # compute total size of the antenna array in x, y, and z directions
        antenna_array_size = self._antenna_nominal_size + antenna_size

        new_sim_bounds = [
            antenna_center + self._antenna_nominal_center - antenna_array_size / 2 - buffer_min,
            antenna_center + self._antenna_nominal_center + antenna_array_size / 2 + buffer_max,
        ]

        # compute the total size of the simulation domain
        new_sim_size = new_sim_bounds[1] - new_sim_bounds[0]
        new_sim_center = 0.5 * (new_sim_bounds[0] + new_sim_bounds[1])

        # duplicate structures, sources, lumped elements, and override structures
        array_structures = self._duplicate_structures(
            structures=simulation.structures,
            new_sim_bounds=new_sim_bounds,
            old_sim_bounds=sim_bounds,
        )
        array_sources, array_lumped_elements = self._duplicate_sources(
            sources=simulation.sources,
            lumped_elements=simulation.lumped_elements,
            old_sim_bounds=sim_bounds,
            new_sim_bounds=new_sim_bounds,
        )
        array_monitors = self._expand_monitors(
            monitors=simulation.monitors,
            antenna_bounds=antenna_bounds,
            new_sim_bounds=new_sim_bounds,
            old_sim_bounds=sim_bounds,
        )
        array_grid_spec = self._duplicate_grid_specs(
            grid_spec=simulation.grid_spec, old_sim_bounds=sim_bounds, new_sim_bounds=new_sim_bounds
        )

        new_sim = simulation.updated_copy(
            center=tuple(new_sim_center),
            size=tuple(new_sim_size),
            structures=array_structures,
            monitors=array_monitors,
            sources=array_sources,
            lumped_elements=array_lumped_elements,
            grid_spec=array_grid_spec,
        )

        return new_sim

    @abstractmethod
    def array_factor(
        self,
        theta: Union[float, ArrayLike],
        phi: Union[float, ArrayLike],
        frequency: Union[NonNegativeFloat, ArrayLike],
    ) -> ArrayLike:
        """
        Compute the array factor for an antenna array.

        Parameters:
        -----------
        theta : Union[float, ArrayLike]
            Observation angles in the elevation plane (in radians).
        phi : Union[float, ArrayLike]
            Observation angles in the azimuth plane (in radians).
        frequency : Union[NonNegativeFloat, ArrayLike]
            Signal frequency (in Hz).

        Returns:
        --------
        ArrayLike
            Array factor values for each combination of theta and phi.
        """

    def monitor_data_from_array_factor(
        self,
        monitor_data: AbstractFieldProjectionData,
        new_monitor: AbstractFieldProjectionMonitor = None,
    ) -> AbstractFieldProjectionData:
        """Apply the array factor to the monitor data of a single antenna.

        Parameters:
        ----------
        monitor_data : AbstractFieldProjectionData
            The monitor data of a single antenna.
        new_monitor : AbstractFieldProjectionMonitor = None
            The new monitor to be used in the resulting data.

        Returns:
        --------
        AbstractFieldProjectionData
            The monitor data of the antenna array.
        """

        # Get spherical coordinates
        r, theta, phi = list(monitor_data.coords_spherical.values())
        freqs = monitor_data.f
        coords_shape = list(np.shape(r))
        coords_shape.append(len(freqs))

        # Compute the array factor
        af = self.array_factor(theta.ravel(), phi.ravel(), freqs, monitor_data.medium)
        af = np.reshape(af, coords_shape)

        update_dict = {}

        # Apply the array factor to the monitor data
        for key, field in monitor_data.field_components.items():
            update_dict[key] = field.copy() * af

        if new_monitor is not None:
            monitor = new_monitor
        else:
            monitor = monitor_data.monitor

        update_dict["monitor"] = monitor
        update_dict["projection_surfaces"] = monitor.projection_surfaces

        if isinstance(monitor_data, DirectivityData):
            # Attempt to recompute flux
            try:
                update_dict["flux"] = monitor_data.flux_from_projected_fields()
            except ValueError as e:
                log.warning(
                    "Could not recalculate flux for a 'DirectivityData' due to the following "
                    f"reason: {e} This monitor will not be included in the resulting data."
                )
                return None

        # Create a new monitor data with the updated fields
        new_mnt_data = monitor_data.updated_copy(
            **update_dict,
        )

        return new_mnt_data

    def simulation_data_from_array_factor(
        self,
        antenna_data: SimulationData,
    ) -> SimulationData:
        """
        Computes the far-field data of a rectangular antenna array based on the far-field data of
        a single antenna. Additionaly, it automatically converts a single antenna simulation setup
        into the corresponding antenna array simulation setup.

        Note that any near-field monitor data will be ignored.

        Parameters:
        ----------
        antenna_data : SimulationData
            The far-field data of a single antenna.

        Returns:
        --------
        SimulationData
            The far-field data of the antenna array.
        """

        # create an expanded simulation for reference
        sim_array = self.make_antenna_array(antenna_data.simulation)

        # names of transferred monitors
        mnt_dict = {mnt.name: mnt for mnt in sim_array.monitors}

        # process far field data
        data_array = []
        good_monitors = []
        for mnt_data in antenna_data.data:
            mnt_name = mnt_data.monitor.name
            if mnt_name in mnt_dict:
                array_mnt_data = self.monitor_data_from_array_factor(
                    monitor_data=mnt_data,
                    new_monitor=mnt_dict[mnt_name],
                )
                if array_mnt_data is not None:
                    good_monitors.append(mnt_dict[mnt_name])
                    data_array.append(array_mnt_data)

        return SimulationData(
            simulation=sim_array.updated_copy(monitors=good_monitors), data=data_array
        )


class RectangularAntennaArrayCalculator(AbstractAntennaArrayCalculator):
    """This class provides methods to calculate the array factor and far-field radiation patterns
    for rectangular phased antenna arrays. It handles arrays with arbitrary size, spacing,
    phase shifts, and amplitude tapering in x, y, and z directions.

    The array factor is calculated using the standard array factor formula for rectangular arrays,
    which accounts for the spatial distribution of antennas and their relative phases and amplitudes.
    This can be used to analyze beam steering, sidelobe levels, and other array characteristics.

    In addition, this class provides a convenience method to create an antenna array simulation
    from a single antenna simulation. This can be used to compute the behavior (near-field and/or
    far-field) of the full antenna array directly without any approximations. Such a simulation setup
    can be obtained:
    - by directly calling the `make_antenna_array` function, or
    - by accessing the field `.simulation` of the `SimulationData` object returned by the
      `simulation_data_from_array_factor` method.

    Example:
    --------
    >>> array_calculator = RectangularAntennaArrayCalculator(
    ...    array_size=(3, 4, 5),
    ...    spacings=(0.5, 0.5, 0.5),
    ...    phase_shifts=(0, 0, 0),
    ... )
    """

    array_size: Tuple[PositiveInt, PositiveInt, PositiveInt] = pd.Field(
        title="Array Size",
        description="Number of antennas along x, y, and z directions.",
    )

    spacings: Tuple[NonNegativeFloat, NonNegativeFloat, NonNegativeFloat] = pd.Field(
        title="Antenna Spacings",
        description="Center-to-center spacings between antennas along x, y, and z directions.",
    )

    phase_shifts: Tuple[float, float, float] = pd.Field(
        (0, 0, 0),
        title="Phase Shifts",
        description="Phase-shifts between antennas along x, y, and z directions.",
    )

    amp_multipliers: Tuple[Optional[ArrayLike], Optional[ArrayLike], Optional[ArrayLike]] = (
        pd.Field(
            (None, None, None),
            title="Amplitude Multipliers",
            description="Amplitude multipliers spatially distributed along x, y, and z directions.",
        )
    )

    @pd.validator("amp_multipliers", pre=True, always=True)
    @skip_if_fields_missing(["array_size"])
    def _check_amp_multipliers(cls, val, values):
        """Check that the length of the amplitude multipliers is equal to the array size along each dimension."""
        array_size = values.get("array_size")
        if len(val) != 3:
            raise ValueError("'amp_multipliers' must have 3 elements.")
        if val[0] is not None and len(val[0]) != array_size[0]:
            raise ValueError(
                f"'amp_multipliers' has length of {len(val[0])} along the x direction, but the array size is {array_size[0]}."
            )
        if val[1] is not None and len(val[1]) != array_size[1]:
            raise ValueError(
                f"'amp_multipliers' has length of {len(val[1])} along the y direction, but the array size is {array_size[1]}."
            )
        if val[2] is not None and len(val[2]) != array_size[2]:
            raise ValueError(
                f"'amp_multipliers' has length of {len(val[2])} along the z direction, but the array size is {array_size[2]}."
            )
        return val

    @property
    def _antenna_locations(self) -> ArrayLike:
        """Locations of antennas' centers in an array."""

        x = (np.arange(self.array_size[0]) - (self.array_size[0] - 1) / 2) * self.spacings[0]
        y = (np.arange(self.array_size[1]) - (self.array_size[1] - 1) / 2) * self.spacings[1]
        z = (np.arange(self.array_size[2]) - (self.array_size[2] - 1) / 2) * self.spacings[2]

        X, Y, Z = np.meshgrid(x, y, z)

        return np.transpose([X.ravel(), Y.ravel(), Z.ravel()])

    @property
    def _antenna_amps(self) -> ArrayLike:
        """Amplitude multipliers of antennas in an array."""

        amps_per_dim = [
            np.ones(size) if multiplier is None else multiplier
            for multiplier, size in zip(self.amp_multipliers, self.array_size)
        ]

        amps_grid = np.meshgrid(*amps_per_dim)

        return np.ravel(amps_grid[0] * amps_grid[1] * amps_grid[2])

    @property
    def _antenna_phases(self) -> ArrayLike:
        """Phase shifts of antennas in an array."""

        phase_shifts_per_dim = [
            np.arange(self.array_size[dim]) * self.phase_shifts[dim] for dim in range(3)
        ]

        phase_shifts_grid = np.meshgrid(*phase_shifts_per_dim)

        return np.ravel(sum(p for p in phase_shifts_grid))

    @property
    def _extend_dims(self) -> Tuple[Axis, ...]:
        """Dimensions along which antennas will be duplicated."""
        return [ind for ind, size in enumerate(self.array_size) if size > 1]

    def array_factor(
        self,
        theta: Union[float, ArrayLike],
        phi: Union[float, ArrayLike],
        frequency: Union[NonNegativeFloat, ArrayLike],
        medium: MediumType3D = Medium(),
    ) -> ArrayLike:
        """
        Compute the array factor for a 3D antenna array.

        Parameters:
        -----------
        theta : Union[float, ArrayLike]
            Observation angles in the elevation plane (in radians).
        phi : Union[float, ArrayLike]
            Observation angles in the azimuth plane (in radians).
        frequency : Union[NonNegativeFloat, ArrayLike]
            Signal frequency (in Hz).

        Returns:
        --------
        ArrayLike
            Array factor values for each combination of theta and phi.
        """

        # Convert all inputs to numpy arrays
        theta_array = np.atleast_1d(theta)
        phi_array = np.atleast_1d(phi)

        # ensure that theta and phi have the same length
        if len(theta_array) != len(phi_array):
            raise ValueError("'theta' and 'phi' must have the same length")

        # reshape inputs for easier broadcasting
        theta_array = np.reshape(theta_array, (len(theta_array), 1))
        phi_array = np.reshape(phi_array, (len(phi_array), 1))
        f_array = np.reshape(frequency, (1, len(np.atleast_1d(frequency))))

        eps_complex = np.array(medium.eps_model(f_array))

        wavelength = C_0 / f_array / np.sqrt(eps_complex)
        k = 2 * np.pi / wavelength  # Wavenumber

        # Calculate the phase shift in the x, y, and z directions
        psi_x = (
            k * self.spacings[0] * np.sin(theta_array) * np.cos(phi_array) - self.phase_shifts[0]
        )
        psi_y = (
            k * self.spacings[1] * np.sin(theta_array) * np.sin(phi_array) - self.phase_shifts[1]
        )
        psi_z = k * self.spacings[2] * np.cos(theta_array) - self.phase_shifts[2]

        amp_x = 1.0 if self.amp_multipliers[0] is None else self.amp_multipliers[0][:, None, None]
        amp_y = 1.0 if self.amp_multipliers[1] is None else self.amp_multipliers[1][:, None, None]
        amp_z = 1.0 if self.amp_multipliers[2] is None else self.amp_multipliers[2][:, None, None]

        # Calculate the array factor in the x, y, and z directions
        af_x = np.sum(
            amp_x
            * np.exp(-1j * np.arange(self.array_size[0])[:, None, None] * psi_x[np.newaxis, :]),
            axis=0,
        )
        af_y = np.sum(
            amp_y
            * np.exp(-1j * np.arange(self.array_size[1])[:, None, None] * psi_y[np.newaxis, :]),
            axis=0,
        )
        af_z = np.sum(
            amp_z
            * np.exp(-1j * np.arange(self.array_size[2])[:, None, None] * psi_z[np.newaxis, :]),
            axis=0,
        )

        # Calculate the overall array factor
        array_factor = af_x * af_y * af_z

        return array_factor



"""Penalty Functions for adjoint plugin."""

from abc import ABC, abstractmethod

import jax.numpy as jnp
import pydantic.v1 as pd

from ....components.base import Tidy3dBaseModel
from ....components.types import ArrayFloat2D
from ....constants import MICROMETER
from ....log import log
from .filter import BinaryProjector, ConicFilter

# Radius of Curvature Calculation


def is_jax_object(arr) -> bool:
    """Test whether an object is a `jnp.ndarray` or an iterable containing them."""
    if isinstance(arr, jnp.ndarray):
        return True
    if isinstance(arr, (list, tuple)):
        return is_jax_object(arr[0])
    return False


class Penalty(Tidy3dBaseModel, ABC):
    """Abstract penalty class. Initializes with parameters and .evaluate() on a design."""

    @abstractmethod
    def evaluate(self) -> float:
        """Evaluate the penalty on supplied values."""


class RadiusPenalty(Penalty):
    """Computes a penalty for small radius of curvature determined by a fit of points in a 2D plane.

    Note
    ----
    .. math::

        p(r) = \\frac{\\mathrm{exp}(-\\kappa(r - r_{min}))}{1 + \\mathrm{exp}(-\\kappa(r - r_{min}))}

    Note
    ----
    This formula was described by A. Micheals et al.
    "Leveraging continuous material averaging for inverse electromagnetic design",
    Optics Express (2018).

    """

    min_radius: float = pd.Field(
        0.150,
        title="Minimum Radius",
        description="Radius of curvature value below which the penalty ramps to its maximum value.",
        units=MICROMETER,
    )

    alpha: float = pd.Field(
        1.0,
        title="Alpha",
        description="Parameter controlling the strength of the penalty.",
    )

    kappa: float = pd.Field(
        10.0,
        title="Kappa",
        description="Parameter controlling the steepness of the penalty evaluation.",
        units="1/" + MICROMETER,
    )

    wrap: bool = pd.Field(
        False,
        title="Wrap",
        description="Whether to consider the first set of points as connected to the last.",
    )

    def evaluate(self, points: ArrayFloat2D) -> float:
        """Get the average penalty as a function of supplied (x, y) points by
        fitting a spline to the curve and evaluating local radius of curvature compared to a
        desired minimum value. If ``wrap``, it is assumed that the points wrap around to form a
        closed geometry instead of an isolated line segment."""

        if not is_jax_object(points):
            log.warning(
                "The points passed to 'RadiusPenalty.evaluate()' are not a 'jax' array. "
                "If passing the 'JaxPolySlab.vertices' field directly, note that the "
                "derivative information for this field "
                "is no longer traced by jax as of "
                "version '2.7'. "
                "The derivative information is contained in 'JaxPolySlab.vertices_jax'. "
                "Therefore, we recommend changing your code to either pass that field or pass "
                "the output of the parameterization functions directly, eg. "
                "'penalty.evaluate(make_vertices(params))'."
            )

        def quad_fit(p0, pc, p2):
            """Quadratic bezier fit (and derivatives) for three points.
            (x(t), y(t)) = P(t) = P0*t^2 + P1*2*t*(1-t) + P2*(1-t)^2
             t in [0, 1]
            """

            # ensure curve goes through (x1, y1) at t=0.5
            p1 = 2 * pc - p0 / 2 - p2 / 2

            def p(t):
                """Bezier curve parameterization."""
                term0 = (1 - t) ** 2 * (p0 - p1)
                term1 = p1
                term2 = t**2 * (p2 - p1)
                return term0 + term1 + term2

            def d_p(t):
                """First derivative function."""
                d_term0 = 2 * (1 - t) * (p1 - p0)
                d_term2 = 2 * t * (p2 - p1)
                return d_term0 + d_term2

            def d2_p(t):
                """Second derivative function."""
                d2_term0 = 2 * p0
                d2_term1 = -4 * p1
                d2_term2 = 2 * p2
                return d2_term0 + d2_term1 + d2_term2

            return p, d_p, d2_p

        def get_fit_vals(xs, ys):
            """Get the values of the Bezier curve and its derivatives at t=0.5 along the points."""

            ps = jnp.stack((xs, ys), axis=1)
            p0 = ps[:-2]
            pc = ps[1:-1]
            p2 = ps[2:]

            p, d_p, d_2p = quad_fit(p0, pc, p2)

            ps = p(0.5)
            dps = d_p(0.5)
            d2ps = d_2p(0.5)
            return ps.T, dps.T, d2ps.T

        def get_radii_curvature(xs, ys):
            """Get the radii of curvature at each (internal) point along the set of points."""
            _, dps, d2ps = get_fit_vals(xs, ys)
            xp, yp = dps
            xp2, yp2 = d2ps
            num = (xp**2 + yp**2) ** (3.0 / 2.0)
            den = abs(xp * yp2 - yp * xp2)
            return num / den

        def penalty_fn(radius):
            """Get the penalty for a given radius."""
            arg = self.kappa * (radius - self.min_radius)
            exp_arg = jnp.exp(-arg)
            return self.alpha * (exp_arg / (1 + exp_arg))

        xs, ys = jnp.array(points).T
        rs = get_radii_curvature(xs, ys)

        # return the average penalty over the points
        return jnp.sum(penalty_fn(rs)) / len(rs)


class ErosionDilationPenalty(Penalty):
    """Computes a penalty for erosion / dilation of a parameter map not being unity.
    Accepts a parameter array normalized between 0 and 1.
    Uses filtering and projection methods to erode and dilate the features within this array.
    Measures the change in the array after eroding and dilating (and also dilating and eroding).
    Returns a penalty proportional to the magnitude of this change.
    The amount of change under dilation and erosion is minimized if the structure has large feature
    sizes and large radius of curvature relative to the length scale.

    Note
    ----
    For more details, refer to chapter 4 of Hammond, A., "High-Efficiency Topology Optimization
    for Very Large-Scale Integrated-Photonics Inverse Design" (2022).


    .. image:: ../../_static/img/erosion_dilation.png

    """

    length_scale: pd.NonNegativeFloat = pd.Field(
        ...,
        title="Length Scale",
        description="Length scale of erosion and dilation. "
        "Corresponds to ``radius`` in the :class:`ConicFilter` used for filtering. "
        "The parameter array is dilated and eroded by half of this value with each operation. "
        "Roughly corresponds to the desired minimum feature size and radius of curvature.",
        units=MICROMETER,
    )

    pixel_size: pd.PositiveFloat = pd.Field(
        ...,
        title="Pixel Size",
        description="Size of each pixel in the array (must be the same along all dimensions). "
        "Corresponds to ``design_region_dl`` in the :class:`ConicFilter` used for filtering.",
        units=MICROMETER,
    )

    beta: pd.PositiveFloat = pd.Field(
        100.0,
        title="Projection Beta",
        description="Strength of the ``tanh`` projection. "
        "Corresponds to ``beta`` in the :class:`BinaryProjector. "
        "Higher values correspond to stronger discretization.",
    )

    eta0: pd.PositiveFloat = pd.Field(
        0.5,
        title="Projection Midpoint",
        description="Value between 0 and 1 that sets the projection midpoint. In other words, "
        "for values of ``eta0``, the projected values are halfway between minimum and maximum. "
        "Corresponds to ``eta`` in the :class:`BinaryProjector`.",
    )

    delta_eta: pd.PositiveFloat = pd.Field(
        0.01,
        title="Delta Eta Cutoff",
        description="The binarization threshold for erosion and dilation operations "
        "The thresholds are ``0 + delta_eta`` on the low end and ``1 - delta_eta`` on the high end. "
        "The default value balances binarization with differentiability so we strongly suggest "
        "using it unless there is a good reason to set it differently.",
    )

    def conic_filter(self) -> ConicFilter:
        """:class:`ConicFilter` associated with this object."""
        return ConicFilter(radius=self.length_scale, design_region_dl=self.pixel_size)

    def binary_projector(self, eta: float = None) -> BinaryProjector:
        """:class:`BinaryProjector` associated with this object."""

        if eta is None:
            eta = self.eta0

        return BinaryProjector(eta=eta, beta=self.beta, vmin=0.0, vmax=1.0, strict_binarize=False)

    def tanh_projection(self, x: jnp.ndarray, eta: float = None) -> jnp.ndarray:
        """Project an array ``x`` once using ``self.beta`` and ``self.eta0``."""
        return self.binary_projector(eta=eta).evaluate(x)

    def filter_project(self, x: jnp.ndarray, eta: float = None) -> jnp.ndarray:
        """Filter an array ``x`` using length scale and dL and then apply a projection."""
        filter = self.conic_filter()
        projector = self.binary_projector(eta=eta)

        y = filter.evaluate(x)
        return projector.evaluate(y)

    def evaluate(self, x: jnp.ndarray) -> float:
        """
        Penalty associated with erosion/dilation and dilation/erosion not being identity.
        Accepts a parameter array with values normalized between 0 and 1.
        Penalty value is normalized such that the maximum possible penalty is 1.
        """
        eta_dilate = 0.0 + self.delta_eta
        eta_eroded = 1.0 - self.delta_eta

        def fn_dilate(x):
            return self.filter_project(x, eta=eta_dilate)

        def fn_eroded(x):
            return self.filter_project(x, eta=eta_eroded)

        params_dilate_erode = fn_eroded(fn_dilate(x))
        params_erode_dilate = fn_dilate(fn_eroded(x))

        diff = params_dilate_erode - params_erode_dilate

        # edge case: if all diff == 0, then the gradient of sqrt() and norm() is not defined.
        if jnp.all(diff == 0.0):
            return 0.0

        return jnp.linalg.norm(diff) / jnp.linalg.norm(jnp.ones_like(diff))



from typing import Any

import autograd.numpy as anp
import pydantic.v1 as pd

from .base import Expression
from .types import NumberOrExpression, NumberType


class Function(Expression):
    """
    Base class for mathematical functions in expressions.
    """

    operand: NumberOrExpression = pd.Field(
        ...,
        title="Operand",
        description="The operand for the function.",
    )

    _format: str = "{func}({operand})"

    @pd.validator("operand", pre=True, always=True)
    def validate_operand(cls, v):
        """
        Validate and convert operand to an expression.
        """
        return cls._to_expression(v)

    def __init__(self, operand: NumberOrExpression, **kwargs: dict[str, Any]) -> None:
        """
        Initialize the function with an operand.

        Parameters
        ----------
        operand : NumberOrExpression
            The operand for the function.
        kwargs : dict[str, Any]
            Additional keyword arguments.
        """
        super().__init__(operand=operand, **kwargs)

    def __repr__(self):
        """
        Return a string representation of the function.
        """
        return self._format.format(func=self.type, operand=self.operand)


class Sin(Function):
    """
    Sine function expression.
    """

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        """
        Evaluate the sine function.

        Returns
        -------
        NumberType
            Sine of the input value.
        """
        return anp.sin(self.operand(*args, **kwargs))


class Cos(Function):
    """
    Cosine function expression.
    """

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        """
        Evaluate the cosine function.

        Returns
        -------
        NumberType
            Cosine of the input value.
        """
        return anp.cos(self.operand(*args, **kwargs))


class Tan(Function):
    """
    Tangent function expression.
    """

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        """
        Evaluate the tangent function.

        Returns
        -------
        NumberType
            Tangent of the input value.
        """
        return anp.tan(self.operand(*args, **kwargs))


class Exp(Function):
    """
    Exponential function expression.
    """

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        """
        Evaluate the exponential function.

        Returns
        -------
        NumberType
            Exponential of the input value.
        """
        return anp.exp(self.operand(*args, **kwargs))


class Log(Function):
    """
    Natural logarithm function expression.
    """

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        """
        Evaluate the natural logarithm function.

        Returns
        -------
        NumberType
            Natural logarithm of the input value.
        """
        return anp.log(self.operand(*args, **kwargs))


class Log10(Function):
    """
    Base-10 logarithm function expression.
    """

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        """
        Evaluate the base-10 logarithm function.

        Returns
        -------
        NumberType
            Base-10 logarithm of the input value.
        """
        return anp.log10(self.operand(*args, **kwargs))


class Sqrt(Function):
    """
    Square root function expression.
    """

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        """
        Evaluate the square root function.

        Returns
        -------
        NumberType
            Square root of the input value.
        """
        return anp.sqrt(self.operand(*args, **kwargs))



from abc import ABC, abstractmethod
from typing import Any, Optional, Union

import autograd.numpy as np
import pydantic.v1 as pd
import xarray as xr

from tidy3d.components.monitor import ModeMonitor
from tidy3d.components.types import Direction, FreqArray

from .base import Expression
from .types import NumberType
from .variables import Variable


def generate_validation_data(expr: Expression) -> dict[str, xr.Dataset]:
    """Generate combined dummy simulation data for all metrics in the expression.

    Parameters
    ----------
    expr : Expression
        The expression containing metrics.

    Returns
    -------
    dict[str, xr.Dataset]
        The combined validation data.
    """
    metrics = set(expr.filter(target_type=Metric))
    combined_data = {k: v for metric in metrics for k, v in metric._validation_data.items()}
    return combined_data


class Metric(Variable, ABC):
    """
    Base class for all metrics.

    To subclass Metric, you must implement an evaluate() method that takes a SimulationData
    object and returns a scalar value.
    """

    @property
    @abstractmethod
    def _validation_data(self) -> Any:
        """Return dummy data for this metric."""

    def __repr__(self) -> str:
        return f'{self.type}("{self.monitor_name}")'


class ModeAmp(Metric):
    """
    Metric for calculating the mode coefficient from a ModeMonitor.

    Examples
    --------
    >>> import tidy3d as td
    >>> monitor = td.ModeMonitor(size=(1, 1, 0), freqs=[2e14], mode_spec=td.ModeSpec(), name="monitor1")
    >>> mode_coeff = ModeAmp.from_mode_monitor(monitor)
    >>> expr = abs(mode_coeff) ** 2
    >>> print(expr)
    (abs(ModeAmp("monitor1")) ** 2)
    """

    monitor_name: str = pd.Field(
        ...,
        title="Monitor Name",
        description="The name of the mode monitor. This needs to match the name of the monitor in the simulation.",
    )
    f: Optional[Union[float, FreqArray]] = pd.Field(  # type: ignore
        None,
        title="Frequency Array",
        description="The frequency array. If None, all frequencies in the monitor will be used.",
        alias="freqs",
    )
    direction: Direction = pd.Field(
        "+",
        title="Direction",
        description="The direction of propagation of the mode.",
    )
    mode_index: pd.NonNegativeInt = pd.Field(
        0,
        title="Mode Index",
        description="The index of the mode.",
    )

    @classmethod
    def from_mode_monitor(
        cls, monitor: ModeMonitor, mode_index: int = 0, direction: Direction = "+"
    ):
        return cls(
            monitor_name=monitor.name, f=monitor.freqs, mode_index=mode_index, direction=direction
        )

    @property
    def _validation_data(self) -> Any:
        """Return dummy data for this metric (complex array of mode amplitudes)."""
        f = np.atleast_1d(self.f).tolist() if self.f is not None else [1.0]
        amps_data = np.random.rand(len(f)) + 1j * np.random.rand(len(f))
        amps = xr.DataArray(
            amps_data.reshape(1, 1, -1),
            coords={
                "direction": [self.direction],
                "mode_index": [self.mode_index],
                "f": f,
            },
        )
        return {self.monitor_name: xr.Dataset({"amps": amps})}

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        data = super().evaluate(*args, **kwargs)
        amps = data[self.monitor_name].amps.sel(
            direction=self.direction, mode_index=self.mode_index
        )
        if self.f is not None:
            f = list(self.f) if isinstance(self.f, tuple) else self.f
            amps = amps.sel(f=f, method="nearest")
        return np.squeeze(amps.data)


class ModePower(ModeAmp):
    """
    Metric for calculating the mode power from a ModeMonitor.

    Examples
    --------
    >>> import tidy3d as td
    >>> monitor = td.ModeMonitor(size=(1, 1, 0), freqs=[2e14], mode_spec=td.ModeSpec(), name="monitor1")
    >>> mode_power = ModePower.from_mode_monitor(monitor)
    """

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        amps = super().evaluate(*args, **kwargs)
        return np.abs(amps) ** 2
